@@ -9,7 +9,9 @@
 
 #include <generic/apt/apt.h>
 #include <generic/apt/config_signal.h>
-#include <generic/apt/matchers.h>
+#include <generic/apt/matching/match.h>
+#include <generic/apt/matching/parse.h>
+#include <generic/apt/matching/pattern.h>
 #include <generic/apt/tasks.h>
 
 #include <apt-pkg/algorithms.h>
@@ -19,12 +21,16 @@
 #include <apt-pkg/policy.h>
 #include <apt-pkg/version.h>
 
+#include <stdio.h>
 #include <stdlib.h>
 
 #include <cwidget/generic/util/transcode.h>
 
 namespace cw = cwidget;
 
+using aptitude::cmdline::terminal;
+using boost::shared_ptr;
+
 namespace
 {
   bool cmdline_do_build_depends(const string &pkg,
@@ -36,7 +42,8 @@ namespace
 				pkgset &to_install, pkgset &to_hold,
 				pkgset &to_remove, pkgset &to_purge,
 				int verbose,
-				bool allow_auto)
+				bool allow_auto,
+                                const shared_ptr<terminal> &term)
   {
     aptitude::cmdline::source_package sourcepkg =
       aptitude::cmdline::find_source_package(pkg,
@@ -66,8 +73,8 @@ namespace
 	bool is_conflict = (it->Type == pkgSrcRecords::Parser::BuildConflictIndep ||
 			    it->Type == pkgSrcRecords::Parser::BuildConflict);
 
-	if(!arch_only && (it->Type == pkgSrcRecords::Parser::BuildDependIndep ||
-			  it->Type == pkgSrcRecords::Parser::BuildConflictIndep))
+	if(arch_only && (it->Type == pkgSrcRecords::Parser::BuildDependIndep ||
+			 it->Type == pkgSrcRecords::Parser::BuildConflictIndep))
 	  continue;
 
 	BuildDepList::const_iterator or_group_start = it;
@@ -139,7 +146,8 @@ namespace
 						best_ver.VerStr(),
 						policy,
 						arch_only,
-						allow_auto);
+						allow_auto,
+                                                term);
 			    satisfied = true;
 			  }
 		      }
@@ -173,7 +181,8 @@ namespace
 					    std::string(),
 					    policy,
 					    arch_only,
-					    allow_auto);
+					    allow_auto,
+                                            term);
 		      }
 		  }
 	      }
@@ -184,6 +193,7 @@ namespace
 	    rval = false;
 
 	    std::string build_dep_description = pkgSrcRecords::Parser::BuildDepType(it->Type);;
+	    build_dep_description += ": ";
 	    for(BuildDepList::const_iterator it2 = or_group_start;
 		it2 <= it; ++it2)
 	      {
@@ -199,7 +209,7 @@ namespace
 		    build_dep_description += ")";
 		  }
 	      }
-	    printf(_("Unable to satisfy the build-depends: %s."),
+	    printf(_("Unable to satisfy the build-depends: %s.\n"),
 		   build_dep_description.c_str());
 	  }
       }
@@ -218,7 +228,8 @@ bool cmdline_applyaction(cmdline_pkgaction_type action,
 			 const string &sourcestr,
 			 pkgPolicy &policy,
 			 bool arch_only,
-			 bool allow_auto)
+			 bool allow_auto,
+                         const shared_ptr<terminal> &term)
 {
   // Handle virtual packages.
   if(!pkg.ProvidesList().end())
@@ -274,7 +285,7 @@ bool cmdline_applyaction(cmdline_pkgaction_type action,
 		{
 		  printf(_("\"%s\" is a virtual package provided by:\n"),
 			 pkg.Name());
-		  cmdline_show_pkglist(cands);
+		  cmdline_show_pkglist(cands, term);
 		  printf(_("You must choose one to install.\n"));
 		}
 	      return false;
@@ -295,17 +306,31 @@ bool cmdline_applyaction(cmdline_pkgaction_type action,
   if(action==cmdline_install || action == cmdline_installauto)
     ver=cmdline_find_ver(pkg, source, sourcestr);
 
+  pkgDepCache::StateCache &pkg_state((*apt_cache_file)[pkg]);
+
   switch(action)
     {
     case cmdline_installauto:
     case cmdline_install:
       if(pkg.CurrentVer()!=ver || pkg->CurrentState!=pkgCache::State::Installed)
 	to_install.insert(pkg);
-      else if((*apt_cache_file)[pkg].Keep() && verbose>0)
+      else if(pkg_state.Keep() && verbose>0)
 	printf(_("%s is already installed at the requested version (%s)\n"),
 	       pkg.Name(),
 	       ver.VerStr());
       break;
+    case cmdline_upgrade:
+      if(pkg_state.Status == 1)
+	to_install.insert(pkg);
+      else if(pkg_state.Status > 1)
+	{
+	  if(verbose > 0)
+	    printf(_("%s is not currently installed, so it will not be upgraded.\n"), pkg.Name());
+	}
+      else if(verbose > 0)
+	printf(_("%s is already installed at the latest version, so it will not be upgraded."), pkg.Name());
+      break;
+
     case cmdline_reinstall:
       if(pkg.CurrentVer().end())
 	printf(_("%s is not currently installed, so it will not be reinstalled.\n"), pkg.Name());
@@ -314,13 +339,13 @@ bool cmdline_applyaction(cmdline_pkgaction_type action,
     case cmdline_remove:
       if(!pkg.CurrentVer().end())
 	to_remove.insert(pkg);
-      else if((*apt_cache_file)[pkg].Keep() && verbose>0)
+      else if(pkg_state.Keep() && verbose>0)
 	printf(_("Package %s is not installed, so it will not be removed\n"), pkg.Name());
       break;
     case cmdline_purge:
       if(!pkg.CurrentVer().end() || pkg->CurrentState!=pkgCache::State::ConfigFiles)
 	to_purge.insert(pkg);
-      else if((*apt_cache_file)[pkg].Keep() && verbose>0)
+      else if(pkg_state.Keep() && verbose>0)
 	printf(_("Package %s is not installed, so it will not be removed\n"), pkg.Name());
       break;
     case cmdline_hold:
@@ -340,7 +365,7 @@ bool cmdline_applyaction(cmdline_pkgaction_type action,
 	{
 	  if(pkg.CurrentVer().end())
 	    printf(_("Package %s is not installed, cannot forbid an upgrade\n"), pkg.Name());
-	  else if(!(*apt_cache_file)[pkg].Upgradable())
+	  else if(!pkg_state.Upgradable())
 	    printf(_("Package %s is not upgradable, cannot forbid an upgrade\n"), pkg.Name());
 	}
     default:
@@ -361,6 +386,11 @@ bool cmdline_applyaction(cmdline_pkgaction_type action,
       if(action == cmdline_installauto)
 	(*apt_cache_file)->mark_auto_installed(pkg, true, NULL);
       break;
+    case cmdline_upgrade:
+      if(pkg_state.Status == 1)
+	(*apt_cache_file)->mark_install(pkg, allow_auto && aptcfg->FindB(PACKAGE "::Auto-Install", true),
+					false, NULL);
+      break;
     case cmdline_remove:
       (*apt_cache_file)->mark_delete(pkg, false, false, NULL);
       break;
@@ -374,7 +404,7 @@ bool cmdline_applyaction(cmdline_pkgaction_type action,
       (*apt_cache_file)->mark_keep(pkg, false, false, NULL);
       break;
     case cmdline_unhold:
-      if((*apt_cache_file)[pkg].Keep())
+      if(pkg_state.Keep())
 	(*apt_cache_file)->mark_keep(pkg, false, false, NULL);
       break;
     case cmdline_markauto:
@@ -389,7 +419,7 @@ bool cmdline_applyaction(cmdline_pkgaction_type action,
       else
 	{
 	  pkgCache::VerIterator curver=pkg.CurrentVer();
-	  pkgCache::VerIterator candver=(*apt_cache_file)[pkg].CandidateVerIter(*apt_cache_file);
+	  pkgCache::VerIterator candver = pkg_state.CandidateVerIter(*apt_cache_file);
 	  if(!curver.end() && !candver.end() && curver!=candver)
 	    (*apt_cache_file)->forbid_upgrade(pkg, candver.VerStr(), NULL);
 	}
@@ -404,7 +434,8 @@ bool cmdline_applyaction(cmdline_pkgaction_type action,
 				      to_install, to_hold,
 				      to_remove, to_purge,
 				      verbose,
-				      allow_auto);
+				      allow_auto,
+                                      term);
     default:
       fprintf(stderr, "Internal error: impossible pkgaction type\n");
       abort();
@@ -420,7 +451,8 @@ bool cmdline_applyaction(string s,
 			 pkgset &to_remove, pkgset &to_purge,
 			 int verbose,
 			 pkgPolicy &policy, bool arch_only,
-			 bool allow_auto)
+			 bool allow_auto,
+                         const shared_ptr<terminal> &term)
 {
   using namespace aptitude::matching;
 
@@ -452,7 +484,8 @@ bool cmdline_applyaction(string s,
 				       verbose, source,
 				       sourcestr,
 				       policy, arch_only,
-				       allow_auto) && rval;
+				       allow_auto,
+                                       term) && rval;
 	}
 
       // break out.
@@ -482,7 +515,7 @@ bool cmdline_applyaction(string s,
       return false;
     }
 
-  if(!cmdline_is_search_pattern(package))
+  if(!aptitude::matching::is_pattern(package))
     {
       pkgCache::PkgIterator pkg=(*apt_cache_file)->FindPkg(package.c_str());
       if(pkg.end())
@@ -498,7 +531,8 @@ bool cmdline_applyaction(string s,
 					    to_install, to_hold,
 					    to_remove, to_purge,
 					    verbose,
-					    allow_auto);
+					    allow_auto,
+                                            term);
 
 	  // Maybe they misspelled the package name?
 	  pkgvector possible;
@@ -519,7 +553,7 @@ bool cmdline_applyaction(string s,
 	      else
 		{
 		  printf(_("Couldn't find package \"%s\".  However, the following\npackages contain \"%s\" in their name:\n"), package.c_str(), package.c_str());
-		  cmdline_show_pkglist(possible);
+		  cmdline_show_pkglist(possible, term);
 		}
 	    }
 	  else
@@ -546,7 +580,7 @@ bool cmdline_applyaction(string s,
 	      else
 		{
 		  printf(_("Couldn't find any package matching \"%s\".  However, the following\npackages contain \"%s\" in their description:\n"), package.c_str(), package.c_str());
-		  cmdline_show_pkglist(possible);
+		  cmdline_show_pkglist(possible, term);
 		}
 	    }
 
@@ -557,33 +591,36 @@ bool cmdline_applyaction(string s,
 				 seen_virtual_packages,
 				 to_install, to_hold, to_remove, to_purge,
 				 verbose, source,
-				 sourcestr, policy, arch_only, allow_auto);
+				 sourcestr, policy, arch_only, allow_auto,
+                                 term);
     }
   else
     {
-      pkg_matcher *m = parse_pattern(package.c_str());
-      if(!m)
+      cw::util::ref_ptr<pattern> p(parse(package.c_str()));
+      if(!p.valid())
 	{
 	  _error->DumpErrors();
 	  return false;
 	}
 
-      for(pkgCache::PkgIterator pkg=(*apt_cache_file)->PkgBegin();
-	  !pkg.end(); ++pkg)
+      std::vector<std::pair<pkgCache::PkgIterator, cw::util::ref_ptr<structural_match> > > matches;
+      cw::util::ref_ptr<search_cache> search_info(search_cache::create());
+      search(p, search_info, matches,
+	     *apt_cache_file,
+	     *apt_package_records);
+      for(std::vector<std::pair<pkgCache::PkgIterator, cw::util::ref_ptr<structural_match> > >::const_iterator
+	    it = matches.begin(); it != matches.end(); ++it)
 	{
-	  pkgCache::VerIterator testver;
-
-	  if(apply_matcher(m, pkg, *apt_cache_file, *apt_package_records))
-	    rval = cmdline_applyaction(action, pkg,
-				       seen_virtual_packages,
-				       to_install, to_hold, to_remove, to_purge,
-				       verbose, source,
-				       sourcestr,
-				       policy, arch_only,
-				       allow_auto) && rval;
+	  if(!cmdline_applyaction(action, it->first,
+				  seen_virtual_packages,
+				  to_install, to_hold, to_remove, to_purge,
+				  verbose, source,
+				  sourcestr,
+				  policy, arch_only,
+				  allow_auto,
+                                  term))
+	    rval = false;
 	}
-
-      delete m;
     }
 
   return rval;
@@ -669,7 +706,8 @@ void cmdline_parse_action(string s,
 			  pkgset &to_remove, pkgset &to_purge,
 			  int verbose,
 			  pkgPolicy &policy, bool arch_only,
-			  bool allow_auto)
+			  bool allow_auto,
+                          const boost::shared_ptr<aptitude::cmdline::terminal> &term)
 {
   string::size_type loc=0;
 
@@ -706,7 +744,8 @@ void cmdline_parse_action(string s,
 				      to_install, to_hold,
 				      to_remove, to_purge,
 				      verbose, policy,
-				      arch_only, allow_auto))
+				      arch_only, allow_auto,
+                                      term))
 		return;
 	    }
 	}
