@@ -1,17 +1,46 @@
 // cmdline_action.h                                   -*-c++-*-
 //
-//   Copyright 2004 Daniel Burrows
+// Copyright (C) 2004, 2010 Daniel Burrows
+//
+// This program is free software; you can redistribute it and/or
+// modify it under the terms of the GNU General Public License as
+// published by the Free Software Foundation; either version 2 of the
+// License, or (at your option) any later version.
+//
+// This program is distributed in the hope that it will be useful, but
+// WITHOUT ANY WARRANTY; without even the implied warranty of
+// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+// General Public License for more details.
+//
+// You should have received a copy of the GNU General Public License
+// along with this program; see the file COPYING.  If not, write to
+// the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
+// Boston, MA 02111-1307, USA.
 
 #ifndef CMDLINE_ACTION_H
 #define CMDLINE_ACTION_H
 
+// Local includes:
 #include "cmdline_common.h"
 
+
+// System includes:
+#include <boost/shared_ptr.hpp>
+
+
 /** \file cmdline_action.h
  */
 
 class pkgPolicy;
 
+namespace aptitude
+{
+  namespace cmdline
+  {
+    class terminal;
+  }
+}
+
 /// \todo The command-line state should probably be encapsulated
 /// as an object.
 
@@ -81,7 +110,8 @@ bool cmdline_applyaction(cmdline_pkgaction_type action,
 			 const string &sourcestr,
 			 pkgPolicy &policy,
 			 bool arch_only,
-			 bool allow_auto);
+			 bool allow_auto,
+                         const boost::shared_ptr<aptitude::cmdline::terminal> &term);
 
 /** \brief Apply the given command-line action to the given package,
  *  updating the command-line state appropriately.
@@ -141,7 +171,8 @@ bool cmdline_applyaction(string s,
 			 pkgset &to_remove, pkgset &to_purge,
 			 int verbose,
 			 pkgPolicy &policy, bool arch_only,
-			 bool allow_auto);
+			 bool allow_auto,
+                         const boost::shared_ptr<aptitude::cmdline::terminal> &term);
 
 /** \brief Parses a list of actions and executes them.
  *
@@ -197,6 +228,7 @@ void cmdline_parse_action(string s,
 			  pkgset &to_remove, pkgset &to_purge,
 			  int verbose,
 			  pkgPolicy &policy, bool arch_only,
-			  bool allow_auto);
+			  bool allow_auto,
+                          const boost::shared_ptr<aptitude::cmdline::terminal> &term);
 
 #endif // CMDLINE_ACTION_H
