@@ -1,7 +1,24 @@
 // cmdline_do_action.cc
 //
-//  Copyright 2004 Daniel Burrows
+//  Copyright (C) 2004, 2010 Daniel Burrows
+//
+// This program is free software; you can redistribute it and/or
+// modify it under the terms of the GNU General Public License as
+// published by the Free Software Foundation; either version 2 of the
+// License, or (at your option) any later version.
+//
+// This program is distributed in the hope that it will be useful, but
+// WITHOUT ANY WARRANTY; without even the implied warranty of
+// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+// General Public License for more details.
+//
+// You should have received a copy of the GNU General Public License
+// along with this program; see the file COPYING.  If not, write to
+// the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
+// Boston, MA 02111-1307, USA.
 
+
+// Local includes:
 #include "cmdline_do_action.h"
 
 #include "cmdline_action.h"
@@ -11,6 +28,8 @@
 #include "cmdline_show_broken.h"
 #include "cmdline_simulate.h"
 #include "cmdline_util.h"
+#include "terminal.h"
+#include "text_progress.h"
 
 #include <generic/apt/apt.h>
 #include <generic/apt/config_signal.h>
@@ -18,12 +37,53 @@
 
 #include <aptitude.h>
 
+
+// System includes:
 #include <apt-pkg/algorithms.h>
 #include <apt-pkg/error.h>
+#include <apt-pkg/policy.h>
 #include <apt-pkg/progress.h>
 
+#include <stdio.h>
+
 using namespace std;
 
+using aptitude::cmdline::create_terminal;
+using aptitude::cmdline::create_terminal_locale;
+using aptitude::cmdline::make_text_progress;
+using aptitude::cmdline::terminal;
+using aptitude::cmdline::terminal_locale;
+using boost::shared_ptr;
+
+namespace
+{
+  void run_dpkg_directly(sigc::slot1<pkgPackageManager::OrderResult, int> f,
+			 sigc::slot1<void, pkgPackageManager::OrderResult> k)
+  {
+    k(f(aptcfg->FindI("APT::Status-Fd", -1)));
+  }
+}
+
+/** \brief Used to track whether an upgrade is being performed, and if
+ *  so, what sort of upgrade is being performed.
+ */
+enum upgrade_mode_tp
+  {
+    /** \brief Indicates that no upgrade is being performed. */
+    no_upgrade,
+    /** \brief Indicates that a full-upgrade is being performed: all
+     *  packages will be marked for upgrade by default, and the
+     *  resolver will consider "unsafe" actions.
+     */
+    full_upgrade,
+    /** \brief Indicates that a safe-upgrade is being performed: all
+     *  packages will be marked for upgrade by default, and the
+     *  resolver will not consider "unsafe" actions; automatic
+     *  installation of dependencies is also disabled.
+     */
+    safe_upgrade
+  };
+
 // TODO: add an option to obey sticky states even if it wasn't
 //      explicitly requested.
 //
@@ -32,16 +92,26 @@ using namespace std;
 int cmdline_do_action(int argc, char *argv[],
 		      const char *status_fname, bool simulate,
 		      bool assume_yes, bool download_only, bool fix_broken,
-		      bool showvers, bool showdeps, bool showsize,
+		      bool showvers, bool showdeps,
+		      bool showsize, bool showwhy,
 		      bool visual_preview, bool always_prompt,
+		      resolver_mode_tp resolver_mode, bool safe_resolver_show_actions,
+		      bool no_new_installs, bool no_new_upgrades,
+		      const std::vector<aptitude::cmdline::tag_application> &user_tags,
+		      bool arch_only,
 		      bool queue_only, int verbose)
 {
+  shared_ptr<terminal> term = create_terminal();
+  shared_ptr<terminal_locale> term_locale = create_terminal_locale();
+
   _error->DumpErrors();
 
   cmdline_pkgaction_type default_action=cmdline_install;
-  bool dist_upgrade=false;
 
-  // Parse the action.
+  upgrade_mode_tp upgrade_mode = no_upgrade;
+
+  // Parse the action.  This sets the upgrade mode and (if it was not
+  // overridden by the user) the resolver mode.
   if(!strcasecmp(argv[0], "install"))
     default_action=cmdline_install;
   else if(!strcasecmp(argv[0], "reinstall"))
@@ -49,8 +119,23 @@ int cmdline_do_action(int argc, char *argv[],
   else if(!strcasecmp(argv[0], "full-upgrade") ||
 	  !strcasecmp(argv[0], "dist-upgrade"))
     {
-      default_action=cmdline_install;
-      dist_upgrade=true;
+      default_action = cmdline_upgrade;
+      if(argc == 1)
+	upgrade_mode = full_upgrade;
+      if(resolver_mode == resolver_mode_default)
+	resolver_mode = resolver_mode_full;
+    }
+  else if(!strcasecmp(argv[0], "safe-upgrade") ||
+	  !strcasecmp(argv[0], "upgrade"))
+    {
+      default_action = cmdline_upgrade;
+      // If safe-upgrade is the only argument, we treat this as a full
+      // upgrade; otherwise we treat it as an install-type action that
+      // applies only to the packages listed on the command-line.
+      if(argc == 1)
+	upgrade_mode = safe_upgrade;
+      if(resolver_mode == resolver_mode_default)
+	resolver_mode = resolver_mode_safe;
     }
   else if(!strcasecmp(argv[0], "remove"))
     default_action=cmdline_remove;
@@ -68,6 +153,9 @@ int cmdline_do_action(int argc, char *argv[],
     default_action=cmdline_unmarkauto;
   else if(!strcasecmp(argv[0], "forbid-version"))
     default_action=cmdline_forbid_version;
+  else if(!strcasecmp(argv[0], "build-depends") ||
+	  !strcasecmp(argv[0], "build-dep"))
+    default_action = cmdline_build_depends;
   else
     {
       // Should never happen.
@@ -76,7 +164,10 @@ int cmdline_do_action(int argc, char *argv[],
       return -1;
     }
 
-  OpTextProgress progress(aptcfg->FindI("Quiet", 0));
+  if(resolver_mode == resolver_mode_default)
+    resolver_mode = resolver_mode_full;
+
+  shared_ptr<OpProgress> progress = make_text_progress(false, term, term_locale);
 
   aptcfg->SetNoUser(PACKAGE "::Auto-Upgrade", "false");
 
@@ -85,8 +176,9 @@ int cmdline_do_action(int argc, char *argv[],
   //
   // This way, "aptitude install" will just perform any pending
   // installations.
-  apt_init(&progress, (argc==1 && default_action==cmdline_install &&
-		       !dist_upgrade), status_fname);
+  apt_init(progress.get(),
+           (argc==1 && default_action==cmdline_install &&
+            upgrade_mode == no_upgrade), status_fname);
 
   if(_error->PendingError())
     {
@@ -106,22 +198,19 @@ int cmdline_do_action(int argc, char *argv[],
       return -1;
     }
 
+  pkgPolicy policy(&(*apt_cache_file)->GetCache());
+  ReadPinFile(policy);
+
   pkgset to_upgrade, to_install, to_hold, to_remove, to_purge;
 
-  if(dist_upgrade)
+  if(upgrade_mode == full_upgrade || upgrade_mode == safe_upgrade)
     {
-      // Build to_install to avoid a big printout
-      for(pkgCache::PkgIterator i=(*apt_cache_file)->PkgBegin(); !i.end(); ++i)
-	{
-	  pkgDepCache::StateCache &state=(*apt_cache_file)[i];
+      bool ignore_removed = (upgrade_mode == safe_upgrade);
 
-	  if(!i.CurrentVer().end() &&
-	     state.Upgradable() && !(*apt_cache_file)->is_held(i))
-	    to_upgrade.insert(i);
+      (*apt_cache_file)->get_upgradable(ignore_removed, to_install);
 
-	}
-
-      (*apt_cache_file)->mark_all_upgradable(true, false, NULL);
+      bool use_autoinst = (resolver_mode != resolver_mode_safe);
+      (*apt_cache_file)->mark_all_upgradable(use_autoinst, ignore_removed, NULL);
     }
   /*else if(argc==1 && default_action==cmdline_install)
     {
@@ -187,6 +276,16 @@ int cmdline_do_action(int argc, char *argv[],
 		action = cmdline_keep;
 		target = std::string(argv[i], 0, tmp);
 		break;
+	      case 'D':
+		// "&BD" for installing build depends.
+		if(tmp >= 2 &&
+		   argv[i][tmp - 1] == 'B' &&
+		   argv[i][tmp - 2] == '&')
+		  {
+		    action = cmdline_build_depends;
+		    target = std::string(argv[i], 0, strlen(argv[i]) - 3);
+		  }
+		break;
 	      case 'm':
 	      case 'M':
 		if(tmp > 0 && argv[i][tmp - 1] == '&')
@@ -209,8 +308,12 @@ int cmdline_do_action(int argc, char *argv[],
 
       // If we have auto-install turned on, do a second run over all
       // the packages being installed to blindly resolve their deps.
-      const bool do_autoinstall = aptcfg->FindB(PACKAGE "::Auto-Install", true);
+      // Note that we skip auto-install if the safe resolver is turned
+      // on, on the grounds that auto-install will remove packages for
+      // Conflicts too.
+      const bool do_autoinstall = (resolver_mode != resolver_mode_safe) && aptcfg->FindB(PACKAGE "::Auto-Install", true);
       const int num_passes = do_autoinstall ? 2 : 1;
+      std::set<pkgCache::PkgIterator> seen_virtual_packages;
       for(int pass = 0; pass < num_passes; ++pass)
 	{
 	  // Clear these to avoid undesirable interactions between the
@@ -224,27 +327,47 @@ int cmdline_do_action(int argc, char *argv[],
 	  for(std::vector<action_pair>::const_iterator it = actions.begin();
 	      it != actions.end(); ++it)
 	    {
-	      cmdline_applyaction(it->second, it->first,
+	      cmdline_applyaction(it->second, seen_virtual_packages, it->first,
 				  to_install, to_hold, to_remove, to_purge,
-				  verbose, pass > 0);
+				  verbose, policy, arch_only, pass > 0,
+                                  term);
 	    }
 	}
     }
   }
 
+  if(resolver_mode == resolver_mode_safe)
+    {
+      if(!aptitude::cmdline::safe_resolve_deps(verbose,
+                                               no_new_installs,
+                                               no_new_upgrades,
+                                               safe_resolver_show_actions,
+                                               term))
+	{
+	  fprintf(stderr, _("Unable to safely resolve dependencies, try running with --full-resolver.\n"));
+	  return -1;
+	}
+    }
+
   if(visual_preview)
     {
       ui_preview();
       return 0;
     }
   else if(simulate)
-    return cmdline_simulate(dist_upgrade, to_install, to_hold, to_remove, to_purge,
-			    showvers, showdeps, showsize,
+    return cmdline_simulate(upgrade_mode != no_upgrade,
+			    to_install, to_hold, to_remove, to_purge,
+			    showvers, showdeps,
+			    showsize, showwhy,
 			    always_prompt, verbose, assume_yes,
-			    !fix_broken);
+			    !fix_broken,
+			    policy, arch_only,
+                            term);
   else if(queue_only)
     {
-      if(!(*apt_cache_file)->save_selection_list(progress))
+      aptitude::cmdline::apply_user_tags(user_tags);
+
+      if(!(*apt_cache_file)->save_selection_list(*progress))
 	{
 	  _error->DumpErrors();
 	  return -1;
@@ -254,20 +377,25 @@ int cmdline_do_action(int argc, char *argv[],
     }
   else
     {
-      if(!cmdline_do_prompt(dist_upgrade,
+      if(!cmdline_do_prompt(upgrade_mode != no_upgrade,
 			    to_install, to_hold, to_remove, to_purge,
-			    showvers, showdeps, showsize,
+			    showvers, showdeps, showsize, showwhy,
 			    always_prompt, verbose, assume_yes,
-			    !fix_broken))
+			    !fix_broken,
+			    policy, arch_only, term))
 	{
 	  printf(_("Abort.\n"));
 	  return 0;
 	}
 
-      download_install_manager m(download_only);
+      aptitude::cmdline::apply_user_tags(user_tags);
+
+      download_install_manager m(download_only,
+				 sigc::ptr_fun(&run_dpkg_directly));
 
       int rval =
-	(cmdline_do_download(&m) == download_manager::success ? 0 : -1);
+	(cmdline_do_download(&m, verbose, term, term_locale)
+         == download_manager::success ? 0 : -1);
 
       if(_error->PendingError())
 	rval = -1;
