@@ -7,13 +7,17 @@
 #include "cmdline_common.h"
 #include "cmdline_progress.h"
 #include "cmdline_util.h"
+#include "terminal.h"
+#include "text_progress.h"
 
 #include <aptitude.h>
 
 #include <generic/apt/apt.h>
 #include <generic/apt/config_signal.h>
 #include <generic/apt/download_signal_log.h>
-#include <generic/apt/matchers.h>
+#include <generic/apt/matching/match.h>
+#include <generic/apt/matching/parse.h>
+#include <generic/apt/matching/pattern.h>
 #include <generic/apt/pkg_acqfile.h>
 
 #include <apt-pkg/acquire.h>
@@ -23,9 +27,19 @@
 
 #include <stdio.h>
 
+using aptitude::cmdline::create_terminal;
+using aptitude::cmdline::create_terminal_locale;
+using aptitude::cmdline::make_text_progress;
+using aptitude::cmdline::terminal;
+using aptitude::cmdline::terminal_locale;
+using boost::shared_ptr;
+
 // Download stuff to the current directory
 int cmdline_download(int argc, char *argv[])
 {
+  shared_ptr<terminal> term = create_terminal();
+  shared_ptr<terminal_locale> term_locale = create_terminal_locale();
+
   if(argc<=1)
     {
       printf(_("download: you must specify at least one package to download\n"));
@@ -34,8 +48,8 @@ int cmdline_download(int argc, char *argv[])
 
   _error->DumpErrors();
 
-  OpTextProgress progress(aptcfg->FindI("Quiet", 0));
-  apt_init(&progress, false);
+  shared_ptr<OpProgress> progress = make_text_progress(false, term, term_locale);
+  apt_init(progress.get(), false);
 
   pkgSourceList list;
   if(!list.ReadMainList())
@@ -45,7 +59,7 @@ int cmdline_download(int argc, char *argv[])
       _error->DumpErrors();
       return -1;
     }
-  pkgAcquire fetcher(gen_cmdline_download_progress());
+  pkgAcquire fetcher(gen_cmdline_download_progress(term));
   string filenames[(*apt_cache_file)->Head().PackageCount];
   string default_release = aptcfg->Find("APT::Default-Release");
 
@@ -64,7 +78,7 @@ int cmdline_download(int argc, char *argv[])
 
       std::vector<pkgCache::PkgIterator> packages;
 
-      if(!cmdline_is_search_pattern(name))
+      if(!aptitude::matching::is_pattern(name))
 	{
 	  pkgCache::PkgIterator pkg=(*apt_cache_file)->FindPkg(name);
 	  if(pkg.end())
@@ -78,19 +92,24 @@ int cmdline_download(int argc, char *argv[])
       else
 	{
 	  using namespace aptitude::matching;
-	  std::auto_ptr<pkg_matcher> m(parse_pattern(name.c_str()));
-	  if(m.get() == NULL)
+	  using cwidget::util::ref_ptr;
+	  ref_ptr<pattern> p(parse(name.c_str()));
+	  if(!p.valid())
 	    {
 	      _error->DumpErrors();
 	      return false;
 	    }
 
-	  for(pkgCache::PkgIterator pkg=(*apt_cache_file)->PkgBegin();
-	      !pkg.end(); ++pkg)
-	    {
-	      if(apply_matcher(m.get(), pkg, *apt_cache_file, *apt_package_records))
-		packages.push_back(pkg);
-	    }
+	  std::vector<std::pair<pkgCache::PkgIterator, ref_ptr<structural_match> > > matches;
+	  ref_ptr<search_cache> search_info(search_cache::create());
+	  search(p, search_info,
+		 matches,
+		 *apt_cache_file,
+		 *apt_package_records);
+
+	  for(std::vector<std::pair<pkgCache::PkgIterator, ref_ptr<structural_match> > >::const_iterator
+		it = matches.begin(); it != matches.end(); ++it)
+	    packages.push_back(it->first);
 
 	  // Maybe there should be a warning here if packages is
 	  // empty?  TODO: think about it again when the string freeze
