@@ -1,7 +1,26 @@
 // cmdline_prompt.cc
 //
+// Copyright (C) 2010 Daniel Burrows
+//
+// This program is free software; you can redistribute it and/or
+// modify it under the terms of the GNU General Public License as
+// published by the Free Software Foundation; either version 2 of the
+// License, or (at your option) any later version.
+//
+// This program is distributed in the hope that it will be useful, but
+// WITHOUT ANY WARRANTY; without even the implied warranty of
+// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+// General Public License for more details.
+//
+// You should have received a copy of the GNU General Public License
+// along with this program; see the file COPYING.  If not, write to
+// the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
+// Boston, MA 02111-1307, USA.
+//
 //   Handles the preview and prompt that's displayed from the command-line.
 
+
+// Local includes:
 #include "cmdline_prompt.h"
 
 #include "cmdline_action.h"
@@ -18,29 +37,32 @@
 #include <generic/apt/config_signal.h>
 #include <generic/apt/download_signal_log.h>
 #include <generic/apt/infer_reason.h>
-#include <generic/apt/matchers.h>
+#include <generic/apt/matching/match.h>
+#include <generic/apt/matching/parse.h>
+#include <generic/apt/matching/pattern.h>
 
 #include <generic/util/util.h>
 
-#include <cwidget/fragment.h>
-#include <cwidget/generic/util/transcode.h>
-#include <cwidget/toplevel.h>
-
+// System includes:
 #include <apt-pkg/algorithms.h>
 #include <apt-pkg/dpkgpm.h>
 #include <apt-pkg/error.h>
 #include <apt-pkg/sourcelist.h>
 #include <apt-pkg/strutl.h>
 
+#include <cwidget/fragment.h>
+#include <cwidget/generic/util/transcode.h>
+#include <cwidget/toplevel.h>
+
 #include <algorithm>
 
 using namespace std;
 namespace cw = cwidget;
 
-string StdinEOFException::errmsg() const
-{
-  return _("Unexpected end-of-file on standard input");
-}
+using aptitude::cmdline::terminal;
+using aptitude::why::make_cmdline_why_callbacks;
+using aptitude::why::why_callbacks;
+using boost::shared_ptr;
 
 struct fetchinfo
 {
@@ -129,8 +151,12 @@ namespace
   };
 
   std::string roots_string(const pkgCache::PkgIterator &pkg,
-			   int verbose)
+			   int verbose,
+                           const shared_ptr<terminal> &term)
   {
+    using namespace aptitude::matching;
+    using cw::util::ref_ptr;
+
     using namespace aptitude::why;
     pkgDepCache::StateCache &state((*apt_cache_file)[pkg]);
 
@@ -141,8 +167,8 @@ namespace
       return "";
 
     target t(state.Install() ? target::Install(pkg) : target::Remove(pkg));
-    std::vector<aptitude::matching::pkg_matcher *> leaves;
-    leaves.push_back(aptitude::matching::parse_pattern("?not(?automatic)"));
+    std::vector<ref_ptr<pattern> > leaves;
+    leaves.push_back(parse("?not(?automatic)"));
 
     std::vector<std::vector<action> > reasons;
 
@@ -160,6 +186,8 @@ namespace
 				   search_params::Recommends,
 				   true));
 
+    const boost::shared_ptr<why_callbacks> callbacks =
+      make_cmdline_why_callbacks(verbose, term);
     for(std::vector<search_params>::const_iterator it = params.begin();
 	it != params.end(); ++it)
       {
@@ -167,6 +195,7 @@ namespace
 			      leaves,
 			      *it,
 			      true,
+                              callbacks,
 			      reasons))
 	  break;
       }
@@ -174,142 +203,33 @@ namespace
     if(reasons.size() == 0)
       return "";
 
-    if(verbose == 0)
-      {
-	std::set<std::string> root_names;
-	for(std::vector<std::vector<action> >::const_iterator it =
-	      reasons.begin(); it != reasons.end(); ++it)
-	  {
-	    // Shouldn't happen, but deal anyway.
-	    if(it->empty())
-	      continue; // Generate an internal error here?
-
-	    const action &act(it->front());
-
-	    // This can happen in the case of an impure virtual
-	    // package.  e.g., A is manually installed and B provides
-	    // A.
-	    if(act.get_dep().end())
-	      continue;
-
-	    if(!act.get_dep().end())
-	      root_names.insert(act.get_dep().ParentPkg().Name());
-	  }
-
-	if(root_names.empty())
-	  return "";
-
-	std::string rval;
-
-	bool first = true;
-	for(std::set<std::string>::const_iterator it = root_names.begin();
-	    it != root_names.end(); ++it)
-	  {
-	    if(first)
-	      first = false;
-	    else
-	      rval += ", ";
-
-	    rval += *it;
-	  }
-
-	// ForTranslators: %s is replaced with a comma-delimited list
-	// of package names.
-	return ssprintf(ngettext("(for %s)",
-				 "(for %s)",
-				 root_names.size()),
-			rval.c_str());
-      }
+    std::vector<std::string> reason_strings;
+    roots_string_mode mode;
+    if(verbose > 1)
+      mode = show_chain_with_versions;
+    else if(verbose > 0)
+      mode = show_chain;
     else
+      mode = show_requiring_packages;
+    summarize_reasons(reasons, mode, reason_strings);
+
+    std::string rval = "(";
+    if(mode == show_requiring_packages)
+      rval += "for ";
+    bool first = true;
+    for(std::vector<std::string>::const_iterator it = reason_strings.begin();
+	it != reason_strings.end(); ++it)
       {
-	// If we're being verbose, display whole chains leading to
-	// each target.
+	if(first)
+	  first = false;
+	else
+	  rval += ", ";
 
-	if(reasons.empty())
-	  return "";
-
-	std::string rval;
-	bool first = true;
-	std::sort(reasons.begin(), reasons.end(), compare_first_action());
-	rval += "(";
-	for(std::vector<std::vector<action> >::const_iterator it =
-	      reasons.begin(); it != reasons.end(); ++it)
-	
-	  {
-	    if(it->empty())
-	      continue;
-
-	    if(first)
-	      first = false;
-	    else
-	      rval += ", ";
-
-	    bool first_action = true;
-	    for(std::vector<action>::const_iterator aIt = it->begin();
-		aIt != it->end(); ++aIt)
-	      {
-		if(!first_action)
-		  rval += " ";
-
-		if(!aIt->get_dep().end())
-		  {
-		    const pkgCache::DepIterator &dep(aIt->get_dep());
-
-		    if(first_action)
-		      {
-			rval += const_cast<pkgCache::DepIterator &>(dep).ParentPkg().Name();
-			rval += " ";
-		      }
-
-		    std::string dep_type = const_cast<pkgCache::DepIterator &>(dep).DepType();
-		    rval += cw::util::transcode(cw::util::transcode(dep_type).substr(0, 1));
-		    rval += ": ";
-
-		    rval += const_cast<pkgCache::DepIterator &>(dep).TargetPkg().Name();
-
-		    // Display versioned deps if we're really being
-		    // verbose.
-		    if(verbose > 1 && ((dep->CompareOp & ~pkgCache::Dep::Or) != pkgCache::Dep::NoOp))
-		      {
-			rval += " (";
-			rval += const_cast<pkgCache::DepIterator &>(dep).CompType();
-			rval += " ";
-			rval += dep.TargetVer();
-			rval += ")";
-		      }
-		  }
-		else
-		  {
-		    const pkgCache::PrvIterator &prv(aIt->get_prv());
-		    eassert(!prv.end());
-
-		    if(first_action)
-		      {
-			rval += const_cast<pkgCache::PrvIterator &>(prv).OwnerPkg().Name();
-		      }
-
-		    rval += " ";
-
-		    rval += cw::util::transcode(cw::util::transcode(_("Provides")).substr(0, 1));
-		    rval += "<- ";
-
-		    rval += const_cast<pkgCache::PrvIterator &>(prv).ParentPkg().Name();
-		    if(verbose > 1 && prv.ProvideVersion() != NULL)
-		      {
-			rval += " (";
-			rval += prv.ProvideVersion();
-			rval += ")";
-		      }
-		  }
-
-		first_action = false;
-	      }
-
-	    rval += ")";
-	  }
-
-	return rval;
+	rval += *it;
       }
+    rval += ")";
+
+    return rval;
   }
 }
 
@@ -337,7 +257,8 @@ static void cmdline_show_instinfo(pkgvector &items,
 				  bool showdeps,
 				  bool showsize,
 				  bool showpurge,
-				  bool showwhy)
+				  bool showwhy,
+                                  const shared_ptr<terminal> &term)
 {
   sort(items.begin(), items.end(), pkg_byname_compare);
   strvector output;
@@ -351,16 +272,16 @@ static void cmdline_show_instinfo(pkgvector &items,
       //aptitudeDepCache::aptitude_state &extstate=(*apt_cache_file)->get_ext_state(*i);
       pkgCache::VerIterator instver=state.InstVerIter(*apt_cache_file);
 
-      // Set to true if this package should get attached deps.
-      bool deps_ok = showdeps;
-
       if(showpurge)
 	{
 	  if(state.Delete() && state.iFlags&pkgDepCache::Purge)
 	    tags.push_back('p');
 	}
 
-      switch(find_pkg_state(*i, *apt_cache_file))
+      if(state.InstBroken())
+        tags.push_back('b');
+
+      switch(find_pkg_state(*i, *apt_cache_file, true))
 	{
 	case pkg_auto_remove:
 	case pkg_auto_install:
@@ -370,11 +291,8 @@ static void cmdline_show_instinfo(pkgvector &items,
 	case pkg_unused_remove:
 	  tags.push_back('u');
 	  break;
-	case pkg_broken:
-	  // Do nothing, but don't clear out deps_ok.
-	  break;
-	default:
-	  deps_ok = false;
+        default:
+          break;
 	}
 
       if(!tags.empty())
@@ -454,7 +372,7 @@ static void cmdline_show_instinfo(pkgvector &items,
 
       if(showwhy)
 	{
-	  std::string whystring(roots_string(*i, verbose));
+	  std::string whystring(roots_string(*i, verbose, term));
 	  if(!whystring.empty())
 	    {
 	      s += " ";
@@ -468,91 +386,7 @@ static void cmdline_show_instinfo(pkgvector &items,
       output.push_back(s);
     }
 
-  cmdline_show_stringlist(output);
-}
-
-// Shows broken dependencies for a single package
-
-static void show_broken_deps(pkgCache::PkgIterator pkg)
-{
-  unsigned int indent=strlen(pkg.Name())+3;
-  bool is_first_dep=true;
-  pkgCache::VerIterator ver=(*apt_cache_file)[pkg].InstVerIter(*apt_cache_file);
-
-  printf("  %s:", pkg.Name());
-  for(pkgCache::DepIterator dep=ver.DependsList(); !dep.end(); ++dep)
-    {
-      pkgCache::DepIterator first=dep, prev=dep;
-
-      while(dep->CompareOp & pkgCache::Dep::Or)
-	++dep;
-
-      // Yep, it's broken.
-      if(dep.IsCritical() &&
-	 !((*apt_cache_file)[dep]&pkgDepCache::DepGInstall))
-	{		  
-	  bool is_first_of_or=true;
-	  // Iterate over the OR group, print out the information.
-
-	  do
-	    {
-	      if(!is_first_dep)
-		for(unsigned int i=0; i<indent; ++i)
-		  printf(" ");
-
-	      is_first_dep=false;
-
-	      if(!is_first_of_or)
-		for(unsigned int i=0; i<strlen(dep.DepType())+3; ++i)
-		  printf(" ");
-	      else
-		printf(" %s: ", first.DepType());
-
-	      is_first_of_or=false;
-
-	      if(first.TargetVer())
-		printf("%s (%s %s)", first.TargetPkg().Name(), first.CompType(), first.TargetVer());
-	      else
-		printf("%s", first.TargetPkg().Name());
-
-	      // FIXME: handle virtual packages sanely.
-	      pkgCache::PkgIterator target=first.TargetPkg();
-	      // Don't skip real packages which are provided.
-	      if(!target.VersionList().end())
-		{
-		  printf(" ");
-
-		  pkgCache::VerIterator ver=(*apt_cache_file)[target].InstVerIter(*apt_cache_file);
-
-		  if(!ver.end()) // ok, it's installable.
-		    {
-		      if((*apt_cache_file)[target].Install())
-			printf(_("but %s is to be installed."),
-			       ver.VerStr());
-		      else if((*apt_cache_file)[target].Upgradable())
-			printf(_("but %s is installed and it is kept back."),
-			       target.CurrentVer().VerStr());
-		      else
-			printf(_("but %s is installed."),
-			       target.CurrentVer().VerStr());
-		    }
-		  else
-		    printf(_("but it is not installable"));
-		}
-	      else
-		// FIXME: do something sensible here!
-		printf(_(" which is a virtual package."));
-
-	      if(first!=dep)
-		printf(_(" or"));
-
-	      printf("\n");
-
-	      prev=first;
-	      ++first;
-	    } while(prev!=dep);
-	}
-    }
+  cmdline_show_stringlist(output, term);
 }
 
 // Note that not all of these are actually used, but I'm preserving
@@ -600,7 +434,7 @@ string prompt_string(const string &prompt)
  *  fat warning message about them.  Returns false if the user doesn't
  *  want to continue.
  */
-static bool prompt_essential()
+static bool prompt_essential(const shared_ptr<terminal> &term)
 {
   pkgvector todelete, whatsbroken;
   bool ok=true;
@@ -625,7 +459,7 @@ static bool prompt_essential()
       ok=false;
 
       printf(_("The following ESSENTIAL packages will be REMOVED!\n"));
-      cmdline_show_pkglist(todelete);
+      cmdline_show_pkglist(todelete, term);
       printf("\n");
     }
 
@@ -669,7 +503,7 @@ static bool prompt_essential()
  *
  *  \return true if everything is OK or the user overrode the warning.
  */
-static bool prompt_trust()
+static bool prompt_trust(const shared_ptr<terminal> &term)
 {
   pkgvector untrusted;
 
@@ -696,7 +530,7 @@ static bool prompt_trust()
 	       "You should only proceed with the installation if you are certain that\n"
 	       "this is what you want to do.\n\n"));
 
-      cmdline_show_pkglist(untrusted);
+      cmdline_show_pkglist(untrusted, term);
 
       printf("\n");
 
@@ -718,7 +552,18 @@ static bool prompt_trust()
 	}
 
 
-      const string okstr=_("Yes"), abortstr=_("No");
+      // ForTranslators: This string is a confirmation message, which
+      // users (especially CJK users) should be able to input without
+      // input methods.  Please include nothing but ASCII characters.
+      // The text preceding the pipe character (|) will be ignored and
+      // can be removed from your translation.
+      const string okstr    = P_("Go ahead and ignore the warning|Yes");
+      // ForTranslators: This string is a confirmation message, which
+      // users (especially CJK users) should be able to input without
+      // input methods.  Please include nothing but ASCII characters.
+      // The text preceding the pipe character (|) will be ignored and
+      // can be removed from your translation.
+      const string abortstr = P_("Abort instead of overriding the warning|No");
 
       // These strings are used to compare in a translation-invariant
       // way, so that "yes" and "no" are always valid inputs; if the
@@ -768,10 +613,10 @@ bool cmdline_show_preview(bool as_upgrade, pkgset &to_install,
 			  pkgset &to_hold, pkgset &to_remove,
 			  bool showvers, bool showdeps,
 			  bool showsize, bool showwhy,
-			  int verbose)
+			  int verbose,
+                          const shared_ptr<terminal> &term)
 {
   const int quiet = aptcfg->FindI("Quiet", 0);
-  bool all_empty=true;
 
   pkgvector lists[num_pkg_action_states];
   pkgvector recommended, suggested;
@@ -791,7 +636,7 @@ bool cmdline_show_preview(bool as_upgrade, pkgset &to_install,
 	      ((*apt_cache_file)[pkg].iFlags & pkgDepCache::ReInstall))
 	++ReInstall;
 
-      pkg_action_state state=find_pkg_state(pkg, *apt_cache_file);
+      pkg_action_state state=find_pkg_state(pkg, *apt_cache_file, true);
 
       switch(state)
 	{
@@ -847,8 +692,6 @@ bool cmdline_show_preview(bool as_upgrade, pkgset &to_install,
     {
       if(!lists[i].empty())
 	{
-	  all_empty=false;
-
 	  printf("%s\n", _(cmdline_action_descriptions[i]));
 	  cmdline_show_instinfo(lists[i],
 				verbose,
@@ -856,23 +699,25 @@ bool cmdline_show_preview(bool as_upgrade, pkgset &to_install,
 				i == pkg_remove ||
 				i == pkg_auto_remove ||
 				i == pkg_unused_remove,
-				showwhy);
+				showwhy,
+                                term);
 	}
     }
 
   if(quiet == 0 && !recommended.empty())
     {
       printf(_("The following packages are RECOMMENDED but will NOT be installed:\n"));
-      cmdline_show_instinfo(recommended, verbose, showvers, showdeps, showsize, false, showwhy);
+      cmdline_show_instinfo(recommended, verbose, showvers, showdeps, showsize, false, showwhy, term);
     }
 
   if(verbose>0 && !suggested.empty())
     {
       printf(_("The following packages are SUGGESTED but will NOT be installed:\n"));
-      cmdline_show_instinfo(suggested, verbose, showvers, showdeps, showsize, false, showwhy);
+      cmdline_show_instinfo(suggested, verbose, showvers, showdeps, showsize, false, showwhy, term);
     }
 
-  if(all_empty)
+  if((*apt_cache_file)->DelCount() == 0 &&
+     (*apt_cache_file)->InstCount() == 0)
     printf(_("No packages will be installed, upgraded, or removed.\n"));
 
   printf(_("%lu packages upgraded, %lu newly installed, "),
@@ -917,7 +762,8 @@ bool cmdline_show_preview(bool as_upgrade, pkgset &to_install,
 }
 
 static void cmdline_parse_show(string response,
-			       int verbose)
+			       int verbose,
+                               const shared_ptr<terminal> &term)
 {
   // assume response[0]=='i'
   std::vector<std::string> packages;
@@ -928,13 +774,13 @@ static void cmdline_parse_show(string response,
   else
     for(std::vector<std::string>::const_iterator it = packages.begin();
 	it != packages.end(); ++it)
-      do_cmdline_show(*it, verbose);
+      do_cmdline_show(*it, verbose, term);
 
   prompt_string(_("Press Return to continue."));
 }
 
 // Erm.  Merge w/ above?
-static void cmdline_parse_changelog(string response)
+static void cmdline_parse_changelog(string response, const shared_ptr<terminal> &term)
 {
   vector<string> packages;
   // assume response[0]=='c'
@@ -943,12 +789,13 @@ static void cmdline_parse_changelog(string response)
   if(packages.empty())
     printf(_("No packages found -- enter the package names on the line after 'c'.\n"));
   else
-    do_cmdline_changelog(packages);
+    do_cmdline_changelog(packages, term);
 
   prompt_string(_("Press Return to continue"));
 }
 
-static void cmdline_parse_why(string response)
+static void cmdline_parse_why(string response,
+                              const shared_ptr<terminal> &term)
 {
   vector<string> arguments;
   // assume response[0]=='w'
@@ -961,10 +808,18 @@ static void cmdline_parse_why(string response)
       bool success;
       string root = arguments.back();
       arguments.pop_back();
-      std::auto_ptr<cw::fragment> frag(do_why(arguments, root, false, false, success));
-      update_screen_width();
+      const shared_ptr<why_callbacks> callbacks =
+        make_cmdline_why_callbacks(0, term);
+      std::auto_ptr<cw::fragment> frag(do_why(arguments, root,
+					      aptitude::why::no_summary,
+					      false, false,
+                                              callbacks,
+                                              success));
       if(frag.get() != NULL)
-	std::cout << frag->layout(screen_width, screen_width, cwidget::style());
+        {
+          const unsigned int screen_width = term->get_screen_width();
+          std::cout << frag->layout(screen_width, screen_width, cwidget::style());
+        }
       _error->DumpErrors();
     }
 }
@@ -974,7 +829,9 @@ static inline cw::fragment *flowindentbox(int i1, int irest, cw::fragment *f)
   return indentbox(i1, irest, flowbox(f));
 }
 
-static void prompt_help(ostream &out, bool show_resolver_key)
+static void prompt_help(ostream &out,
+                        bool show_resolver_key,
+                        const shared_ptr<terminal> &term)
 {
   std::vector<cw::fragment *> fragments;
 
@@ -1046,8 +903,22 @@ static void prompt_help(ostream &out, bool show_resolver_key)
   fragments.push_back(flowindentbox(0, 4,
 				    cw::fragf(_("'&BD' to install the build-dependencies of a package."))));
 
+  fragments.push_back(cw::fragf("\n"));
+  fragments.push_back(cw::flowbox(cw::fragf(_("In the list of actions to be performed, some packages will be followed by one or more characters enclosed in braces; for instance: \"aptitude{u}\".  These characters provide extra information about the package's state, and can include any combination of the following:"))));
+  fragments.push_back(cw::fragf("\n"));
+
+  fragments.push_back(flowindentbox(0, 4,
+				    cw::fragf(_("'a': the package was automatically installed or removed."))));
+  fragments.push_back(flowindentbox(0, 4,
+                                    cw::fragf(_("'b': some of the package's dependencies are violated by the proposed changes."))));
+  fragments.push_back(flowindentbox(0, 4,
+				    cw::fragf(_("'p': the package will be purged in addition to being removed."))));
+  fragments.push_back(flowindentbox(0, 4,
+				    cw::fragf(_("'u': the package is being removed because it is unused."))));
+
   cw::fragment *f = indentbox(2, 2, cw::sequence_fragment(fragments));
 
+  const unsigned int screen_width = term->get_screen_width();
   out << _("Commands:") << endl;
   out << f->layout(screen_width, screen_width, cwidget::style());
   delete f;
@@ -1067,7 +938,8 @@ bool cmdline_do_prompt(bool as_upgrade,
 		       bool assume_yes,
 		       bool force_no_change,
 		       pkgPolicy &policy,
-		       bool arch_only)
+		       bool arch_only,
+                       const shared_ptr<terminal> &term)
 {
   bool exit=false;
   bool rval=true;
@@ -1087,7 +959,8 @@ bool cmdline_do_prompt(bool as_upgrade,
       // If we're only doing what the user asked and it's OK to go
       // ahead, we can break out immediately.
       if(!cmdline_show_preview(true, to_install, to_hold, to_remove,
-			       showvers, showdeps, showsize, showwhy, verbose) &&
+			       showvers, showdeps, showsize, showwhy, verbose,
+                               term) &&
 	 first &&
 	 !always_prompt &&
 	 (*apt_cache_file)->BrokenCount()==0)
@@ -1104,7 +977,8 @@ bool cmdline_do_prompt(bool as_upgrade,
 					  force_no_change,
 					  verbose,
 					  policy,
-					  arch_only))
+					  arch_only,
+                                          term))
 		{
 		case aptitude::cmdline::resolver_success:
 		  break;
@@ -1118,6 +992,19 @@ bool cmdline_do_prompt(bool as_upgrade,
 		  break;
 		}
 
+	      if(!exit)
+		{
+		  // Re-display the preview so the user can see any
+		  // changes the resolver made.
+		  cmdline_show_preview(true, to_install, to_hold, to_remove,
+				       showvers, showdeps, showsize, showwhy,
+				       verbose, term);
+
+		  if((*apt_cache_file)->DelCount() == 0 &&
+		     (*apt_cache_file)->InstCount() == 0)
+		    exit = true;
+		}
+
 	      if(first && assume_yes)
 		{
 		  // If we're supposed to assume "yes", then we actually
@@ -1126,15 +1013,6 @@ bool cmdline_do_prompt(bool as_upgrade,
 		  rval = !have_broken;
 		  exit = true;
 		}
-
-	      if(!exit)
-		{
-		  // Re-display the preview so the user can see any
-		  // changes the resolver made.
-		  cmdline_show_preview(true, to_install, to_hold, to_remove,
-				       showvers, showdeps, showsize, showwhy,
-				       verbose);
-		}
 	    }
 	  else
 	    // The internal resolver is disabled, fall back to manual
@@ -1158,6 +1036,7 @@ bool cmdline_do_prompt(bool as_upgrade,
 	      if(first)
 		{
 		  const std::string msg = _("aptitude failed to find a solution to these dependencies.  You can solve them yourself by hand or type 'n' to quit.");
+                  const unsigned int screen_width = term->get_screen_width();
 		  cw::fragment *f = cw::text_fragment(msg);
 		  cout << f->layout(screen_width,
 				    screen_width,
@@ -1201,11 +1080,12 @@ bool cmdline_do_prompt(bool as_upgrade,
 		  if(have_broken)
 		    {
 		      cw::fragment *f = flowbox(cw::text_fragment(_("Enter a package management command (such as '+ package' to install a package), 'R' to attempt automatic dependency resolution or 'N' to abort.")));
+                      const unsigned int screen_width = term->get_screen_width();
 		      cout << f->layout(screen_width,
 					screen_width,
 					cwidget::style());
 		      delete f;
-		      
+
 		      valid_response = false;
 		    }
 		  else
@@ -1251,13 +1131,13 @@ bool cmdline_do_prompt(bool as_upgrade,
 		    printf(_("\nSize changes will not be shown.\n\n"));
 		  break;
 		case 'I':
-		  cmdline_parse_show(response, verbose);
+		  cmdline_parse_show(response, verbose, term);
 		  break;
 		case 'C':
-		  cmdline_parse_changelog(response);
+		  cmdline_parse_changelog(response, term);
 		  break;
 		case 'W': // should be 'Y' but that's for "yes"
-		  cmdline_parse_why(response);
+		  cmdline_parse_why(response, term);
 		  break;
 		case '+':
 		case '-':
@@ -1273,14 +1153,15 @@ bool cmdline_do_prompt(bool as_upgrade,
 		    cmdline_parse_action(response, seen_virtual_packages,
 					 to_install, to_hold,
 					 to_remove, to_purge, verbose,
-					 policy, arch_only, true);
+					 policy, arch_only, true,
+                                         term);
 		  }
 		  break;
 		case 'E':
 		  ui_preview();
 		case '?':
 		  valid_response=false;
-		  prompt_help(cout, have_broken);
+		  prompt_help(cout, have_broken, term);
 		  break;
 		default:
 		  printf("%s", unknown_key_message.c_str());
@@ -1291,7 +1172,7 @@ bool cmdline_do_prompt(bool as_upgrade,
 	}
 
       // Note: only show the prompt if we're planning to continue.
-      if(rval && (!prompt_essential() || !prompt_trust()))
+      if(rval && (!prompt_essential(term) || !prompt_trust(term)))
 	{
 	  rval=false;
 	  exit=true;
