@@ -1,6 +1,6 @@
 // cmdline_resolver.cc
 //
-//   Copyright (C) 2005-2008 Daniel Burrows
+//   Copyright (C) 2005-2010 Daniel Burrows
 //
 //   This program is free software; you can redistribute it and/or
 //   modify it under the terms of the GNU General Public License as
@@ -17,6 +17,8 @@
 //   the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 //   Boston, MA 02111-1307, USA.
 
+
+// Local includes:
 #include "cmdline_resolver.h"
 
 #include "cmdline_action.h"
@@ -26,6 +28,7 @@
 #include "cmdline_show_broken.h"
 #include "cmdline_spinner.h"
 #include "cmdline_util.h"
+#include "terminal.h"
 
 #include <aptitude.h>
 #include <solution_fragment.h>
@@ -34,12 +37,17 @@
 #include <generic/apt/aptitude_resolver_universe.h>
 #include <generic/apt/config_signal.h>
 #include <generic/apt/resolver_manager.h>
-
 #include <generic/problemresolver/exceptions.h>
 #include <generic/problemresolver/solution.h>
-
 #include <generic/util/util.h>
 
+
+// System includes:
+#include <boost/lexical_cast.hpp>
+#include <boost/make_shared.hpp>
+#include <boost/ref.hpp>
+#include <boost/shared_ptr.hpp>
+
 #include <cwidget/fragment.h>
 #include <cwidget/generic/util/ssprintf.h>
 
@@ -54,31 +62,81 @@
 
 using namespace std;
 
+using aptitude::cmdline::terminal;
+using boost::shared_ptr;
+
 typedef generic_solution<aptitude_universe> aptitude_solution;
+typedef generic_choice<aptitude_universe> choice;
+typedef generic_choice_set<aptitude_universe> choice_set;
 
 namespace cw = cwidget;
 
 /** Generate a cw::fragment describing a solution as an ordered sequence
  *  of actions.
+ *
+ *  \param ids If not NULL, updated with a mapping from the string
+ *  keys identifying choices in the solution to the associated
+ *  choices.
  */
-static cw::fragment *solution_story(const aptitude_solution &s)
+static cw::fragment *solution_story(const aptitude_solution &s,
+				    std::map<std::string, choice> *ids)
 {
-  std::vector<aptitude_solution::action> actions;
-  for(imm::map<aptitude_universe::package, aptitude_solution::action>::const_iterator
-	i = s.get_actions().begin(); i != s.get_actions().end(); ++i)
-    actions.push_back(i->second);
-  sort(actions.begin(), actions.end(), aptitude_solution::action_id_compare());
+  std::vector<choice> choices;
+  for(choice_set::const_iterator it = s.get_choices().begin();
+      it != s.get_choices().end(); ++it)
+    choices.push_back(*it);
+
+  sort(choices.begin(), choices.end(), aptitude_solution::choice_id_compare());
 
 
   vector<cw::fragment *> fragments;
 
-  for(vector<aptitude_solution::action>::const_iterator i = actions.begin();
-      i != actions.end(); ++i)
-    fragments.push_back(cw::fragf("%ls%n -> %F%n",
-			      dep_text(i->d.get_dep()).c_str(),
-			      indentbox(0, 4, action_fragment(*i))));
+  for(vector<choice>::const_iterator i = choices.begin();
+      i != choices.end(); ++i)
+    {
+      fragments.push_back(cw::fragf("%ls",
+				    dep_text(i->get_dep().get_dep()).c_str()));
+      fragments.push_back(cw::fragf(" -> %F%n",
+				    indentbox(0, 4, choice_fragment(*i))));
+    }
 
-  return cw::sequence_fragment(fragments);
+  if(ids == NULL)
+    return cw::sequence_fragment(fragments);
+  else
+    {
+      std::vector<cw::fragment *> id_fragments, flag_fragments;
+
+      for(std::vector<choice>::size_type i = 0; i < choices.size(); ++i)
+	{
+	  std::string key = boost::lexical_cast<std::string>(i + 1);
+	  id_fragments.push_back(NULL);
+	  id_fragments.push_back(cw::fragf("%s)", key.c_str()));
+	  (*ids)[key] = choices[i];
+
+          flag_fragments.push_back(NULL);
+	  flag_fragments.push_back(choice_state_fragment(choices[i]));
+	}
+
+
+      std::vector<cw::fragment_column_entry> columns;
+      columns.push_back(cw::fragment_column_entry(false, true,
+						  1, cw::fragment_column_entry::top,
+						  id_fragments));
+      columns.push_back(cw::fragment_column_entry(false, false,
+						  1, cw::fragment_column_entry::top,
+						  NULL));
+      columns.push_back(cw::fragment_column_entry(false, false,
+						  1, cw::fragment_column_entry::top,
+						  flag_fragments));
+      columns.push_back(cw::fragment_column_entry(false, false,
+						  1, cw::fragment_column_entry::top,
+						  NULL));
+      columns.push_back(cw::fragment_column_entry(false, true,
+						  1, cw::fragment_column_entry::top,
+						  fragments));
+
+      return cw::fragment_columns(columns);
+    }
 }
 
 void cmdline_dump_resolver()
@@ -114,7 +172,8 @@ static void setup_resolver(pkgset &to_install,
   // it.
   resman->reset_resolver();
 
-  resman->set_debug(aptcfg->FindB(PACKAGE "::CmdLine::Resolver-Debug", false));
+  if(aptcfg->FindB(PACKAGE "::CmdLine::Resolver-Debug", false))
+    resman->set_debug(true);
 
   // For all packages that the user listed on the command-line (i.e.,
   // all in to_install, to_hold, to_remove, and to_purge), tell the
@@ -156,7 +215,8 @@ static inline cw::fragment *flowindentbox(int i1, int irest, cw::fragment *f)
   return indentbox(i1, irest, flowbox(f));
 }
 
-static void resolver_help(ostream &out)
+static void resolver_help(ostream &out,
+                          const shared_ptr<terminal> &term)
 {
   cw::fragment *f=indentbox(2, 2,
 			cw::fragf(_("y: %F"
@@ -167,8 +227,9 @@ static void resolver_help(ostream &out)
 				"o: %F"
 				"e: %F"
 				"x: %F"
-				"r pkg ver ...: %F%n"
-				"a pkg ver ...: %F%n"
+				"r (ID|pkg ver) ...: %F%n"
+				"a (ID|pkg ver) ...: %F%n"
+                                "<ID>: %F%n"
 				"<ACTION> pkg... : %F%n"
 				"%F"
 				"%F"
@@ -195,9 +256,11 @@ static void resolver_help(ostream &out)
 			      flowindentbox(0, 3,
 					    cw::fragf(_("abort automatic dependency resolution; resolve dependencies by hand instead"))),
 			      flowindentbox(0, 15,
-					    cw::fragf(_("reject the given package versions; don't display any solutions in which they occur.  Enter UNINST instead of a version to reject removing the package."))),
+					    cw::fragf(_("reject the given package versions; don't display any solutions in which they occur.  Enter UNINST instead of a version to reject removing the package.  ID is the integer printed to the left of the action."))),
 			      flowindentbox(0, 15,
-					    cw::fragf(_("accept the given package versions; display only solutions in which they occur.  Enter UNINST instead of a version to accept removing the package."))),
+					    cw::fragf(_("accept the given package versions; display only solutions in which they occur.  Enter UNINST instead of a version to accept removing the package.  ID is the integer printed to the left of the action."))),
+                              flowindentbox(0, 15,
+                                            cw::fragf(_("display information about the action labeled ID from the solution.  The label is the integer printed to the left of the action."))),
 			      flowindentbox(0, 3,
 					    cw::fragf(_("adjust the state of the listed packages, where ACTION is one of:"))),
 			      flowindentbox(0, 4,
@@ -219,6 +282,7 @@ static void resolver_help(ostream &out)
 			      flowindentbox(0, 3,
 					    cw::fragf(_("Adjustments will cause the current solution to be discarded and recalculated as necessary.")))));
 
+  const unsigned int screen_width = term->get_screen_width();
   out << f->layout(screen_width, screen_width, cwidget::style());
   delete f;
 }
@@ -253,6 +317,7 @@ static pkgCache::VerIterator choose_version(const vector<pkgCache::VerIterator>
 }
 
 static void reject_or_mandate_version(const string &s,
+				      const std::map<std::string, choice> &ids,
 				      bool is_reject)
 {
   istringstream in(s);
@@ -273,105 +338,299 @@ static void reject_or_mandate_version(const string &s,
     {
       in >> pkgname >> ws;
 
-      if(in.eof())
+      choice c; // Initialized below.
+
+      std::map<std::string, choice>::const_iterator found =
+	ids.find(pkgname);
+      if(found != ids.end())
+	c = found->second;
+      else
 	{
-	  cerr << ssprintf(_("Expected a version or \"%s\" after \"%s\""), "UNINST", pkgname.c_str()) << endl;
-	  return;
-	}
-
-      in >> vername >> ws;
-
-      pkgCache::PkgIterator pkg((*apt_cache_file)->FindPkg(pkgname));
-
-      if(pkg.end())
-	{
-	  cerr << ssprintf(_("No such package \"%s\""), pkgname.c_str()) << endl;
-	  continue;
-	}
-
-      aptitude_universe::version ver(pkg,
-				     pkgCache::VerIterator(*apt_cache_file),
-				     *apt_cache_file);
-      if(stringcasecmp(vername, "UNINST") != 0)
-	{
-	  vector<pkgCache::VerIterator> found;
-	  for(pkgCache::VerIterator vi = pkg.VersionList(); !vi.end(); ++vi)
-	    if(vi.VerStr() == vername)
-	      found.push_back(vi);
-
-	  if(found.empty())
+	  if(in.eof())
 	    {
-	      cerr << ssprintf(_("%s has no version named \"%s\""),
-			       pkgname.c_str(), vername.c_str()) << endl;
+	      cerr << ssprintf(_("Expected a version or \"%s\" after \"%s\""), "UNINST", pkgname.c_str()) << endl;
+	      return;
+	    }
+
+	  in >> vername >> ws;
+
+	  pkgCache::PkgIterator pkg((*apt_cache_file)->FindPkg(pkgname));
+
+	  if(pkg.end())
+	    {
+	      cerr << ssprintf(_("No such package \"%s\""), pkgname.c_str()) << endl;
 	      continue;
 	    }
 
-	  ver = aptitude_universe::version(pkg, choose_version(found),
-					   *apt_cache_file);
+	  aptitude_universe::version ver =
+	    aptitude_universe::version::make_removal(pkg,
+						 *apt_cache_file);
+	  if(stringcasecmp(vername, "UNINST") != 0)
+	    {
+	      vector<pkgCache::VerIterator> found;
+	      for(pkgCache::VerIterator vi = pkg.VersionList(); !vi.end(); ++vi)
+		if(vi.VerStr() == vername)
+		  found.push_back(vi);
 
-	  eassert(!ver.get_ver().end());
-	  eassert(ver.get_pkg() == ver.get_ver().ParentPkg());
+	      if(found.empty())
+		{
+		  cerr << ssprintf(_("%s has no version named \"%s\""),
+				   pkgname.c_str(), vername.c_str()) << endl;
+		  continue;
+		}
+
+	      ver = aptitude_universe::version::make_install(choose_version(found),
+							     *apt_cache_file);
+
+	      eassert(!ver.get_ver().end());
+	      eassert(ver.get_pkg() == ver.get_ver().ParentPkg());
+	    }
+
+          c = choice::make_install_version(ver, -1);
 	}
 
-      if(is_reject)
+      switch(c.get_type())
 	{
-	  if(resman->is_rejected(ver))
-	    {
-	      if(ver.get_ver().end())
-		cout << ssprintf(_("Allowing the removal of %s"),
-				 pkgname.c_str()) << endl;
-	      else
-		cout << ssprintf(_("Allowing the installation of %s version %s (%s)"),
-				 pkg.Name(),
-				 ver.get_ver().VerStr(),
-				 archives_text(ver.get_ver()).c_str()) << endl;
+	case choice::install_version:
+	  {
+	    aptitude_resolver_version ver(c.get_ver());
+	    pkgCache::PkgIterator pkg(ver.get_pkg());
+	    if(is_reject)
+	      {
+		if(resman->is_rejected(ver))
+		  {
+		    if(ver.get_ver().end())
+		      cout << ssprintf(_("Allowing the removal of %s"),
+				       pkg.Name()) << endl;
+		    else
+		      cout << ssprintf(_("Allowing the installation of %s version %s (%s)"),
+				       pkg.Name(),
+				       ver.get_ver().VerStr(),
+				       archives_text(ver.get_ver()).c_str()) << endl;
 
-	      resman->unreject_version(ver);
-	    }
-	  else
-	    {
-	      if(ver.get_ver().end())
-		cout << ssprintf(_("Rejecting the removal of %s"),
-				 pkgname.c_str()) << endl;
-	      else
-		cout << ssprintf(_("Rejecting the installation of %s version %s (%s)"),
-				 pkg.Name(),
-				 ver.get_ver().VerStr(),
-				 archives_text(ver.get_ver()).c_str()) << endl;
+		    resman->unreject_version(ver);
+		  }
+		else
+		  {
+		    if(ver.get_ver().end())
+		      cout << ssprintf(_("Rejecting the removal of %s"),
+				       pkg.Name()) << endl;
+		    else
+		      cout << ssprintf(_("Rejecting the installation of %s version %s (%s)"),
+				       pkg.Name(),
+				       ver.get_ver().VerStr(),
+				       archives_text(ver.get_ver()).c_str()) << endl;
 
-	      resman->reject_version(ver);
-	    }
-	}
-      else
-	{
-	  if(resman->is_mandatory(ver))
-	    {
-	      if(ver.get_ver().end())
-		cout << ssprintf(_("No longer requiring the removal of %s"),
-				 pkgname.c_str()) << endl;
-	      else
-		cout << ssprintf(_("No longer requiring the installation of %s version %s (%s)"),
-				 pkg.Name(), ver.get_ver().VerStr(),
-				 archives_text(ver.get_ver()).c_str()) << endl;
+		    resman->reject_version(ver);
+		  }
+	      }
+	    else
+	      {
+		if(resman->is_mandatory(ver))
+		  {
+		    if(ver.get_ver().end())
+		      cout << ssprintf(_("No longer requiring the removal of %s"),
+				       pkg.Name()) << endl;
+		    else
+		      cout << ssprintf(_("No longer requiring the installation of %s version %s (%s)"),
+				       pkg.Name(), ver.get_ver().VerStr(),
+				       archives_text(ver.get_ver()).c_str()) << endl;
 
-	      resman->unmandate_version(ver);
-	    }
-	  else
-	    {
-	      if(ver.get_ver().end())
-		cout << ssprintf(_("Requiring the removal of %s"),
-				 pkgname.c_str()) << endl;
-	      else
-		cout << ssprintf(_("Requiring the installation of %s version %s (%s)"),
-				 pkg.Name(), ver.get_ver().VerStr(),
-				 archives_text(ver.get_ver()).c_str()) << endl;
+		    resman->unmandate_version(ver);
+		  }
+		else
+		  {
+		    if(ver.get_ver().end())
+		      cout << ssprintf(_("Requiring the removal of %s"),
+				       pkg.Name()) << endl;
+		    else
+		      cout << ssprintf(_("Requiring the installation of %s version %s (%s)"),
+				       pkg.Name(), ver.get_ver().VerStr(),
+				       archives_text(ver.get_ver()).c_str()) << endl;
 
-	      resman->mandate_version(ver);
-	    }
+		    resman->mandate_version(ver);
+		  }
+	      }
+	  }
+	  break;
+
+	case choice::break_soft_dep:
+	  {
+	    aptitude_resolver_dep d(c.get_dep());
+
+	    pkgCache::DepIterator start, end;
+	    surrounding_or(d.get_dep(), start, end, d.get_dep().Cache());
+	    std::ostringstream dep_rendering_stream;
+
+	    if(start.end()) // Sanity-check.
+	      dep_rendering_stream << "(??\?)";
+	    else
+	      {
+		dep_rendering_stream << start.ParentPkg().Name()
+				     << " "
+				     << start.ParentVer().VerStr()
+				     << " "
+				     << start.DepType()
+				     << " ";
+		bool first = true;
+
+		while(!start.end() && start != end)
+		  {
+		    if(first)
+		      first = false;
+		    else
+		      dep_rendering_stream << " | ";
+
+		    dep_rendering_stream << start.TargetPkg().Name()
+					 << " "
+					 << start.TargetVer();
+
+		    ++start;
+		  }
+		std::string dep_rendering = dep_rendering_stream.str();
+
+		if(is_reject)
+		  {
+		    if(resman->is_hardened(d))
+		      {
+			cout << ssprintf(_("Allowing this recommendation to be ignored: %s"),
+					 dep_rendering.c_str()) << endl;
+			resman->unharden_dep(d);
+		      }
+		    else
+		      {
+			cout << ssprintf(_("Always obeying this recommendation: %s"),
+					 dep_rendering.c_str()) << endl;
+			resman->harden_dep(d);
+		      }
+		  }
+		else
+		  {
+		    if(resman->is_approved_broken(d))
+		      {
+			cout << ssprintf(_("No longer ignoring this recommendation: %s"),
+					 dep_rendering.c_str()) << endl;
+			resman->unapprove_broken_dep(d);
+		      }
+		    else
+		      {
+			cout << ssprintf(_("Ignoring this recommendation: %s"),
+					 dep_rendering.c_str()) << endl;
+			resman->approve_broken_dep(d);
+		      }
+		  }
+	      }
+	  }
 	}
     }
 }
 
+void cmdline_resolver_show_choice(const choice &c,
+				  const std::string &tag,
+                                  const shared_ptr<terminal> &term)
+{
+  cw::fragment *info_fragment = NULL;
+  bool is_rejected = false;
+  bool is_approved = false;
+
+  switch(c.get_type())
+    {
+    case choice::install_version:
+      {
+        pkgCache::VerIterator disp_ver = c.get_ver().get_ver();
+
+        if(disp_ver.end())
+          disp_ver = c.get_ver().get_pkg().CurrentVer();
+
+        if(disp_ver.end())
+          disp_ver = c.get_ver().get_pkg().VersionList();
+
+        if(disp_ver.end())
+          info_fragment = cw::fragf("Package: %s\n\n",
+                                    c.get_ver().get_pkg().Name());
+        else
+          info_fragment = cw::fragf("%F\n\n",
+                                    version_file_fragment(disp_ver,
+                                                          disp_ver.FileList(),
+                                                          0));
+      }
+
+      is_rejected = resman->is_rejected(c.get_ver());
+      is_approved = resman->is_mandatory(c.get_ver());
+      break;
+
+    case choice::break_soft_dep:
+      info_fragment = cw::text_fragment("");
+      is_rejected = resman->is_hardened(c.get_dep());
+      is_approved = resman->is_approved_broken(c.get_dep());
+      break;
+
+    default:
+      info_fragment = cw::text_fragment("");
+      break;
+    }
+
+  cw::fragment *f =
+    cw::fragf("%F%F%F",
+	      flowbox(cw::fragf("%F: %F\n\n",
+				cw::fragf(_("Action \"%s\""), tag.c_str()),
+				choice_fragment(c))),
+	      info_fragment,
+	      flowbox(cw::fragf("%F\n\n%F%F\n%F\n",
+                                // TODO: maybe this message could be
+                                // improved -- we could probably print
+                                // the dependency in a more
+                                // traditional way and maybe explain
+                                // which part failed?
+                                //
+                                // ForTranslators: the substitution is
+                                // replaced in English with something
+                                // like "Pkg1 depends upon Pkg2".
+                                cw::fragf(_("This action was selected because %ls."),
+                                          dep_text(c.get_dep().get_dep()).c_str()),
+				is_rejected ? cw::fragf("%s\n", _("This action is currently rejected; it will not appear in new solutions.")) :
+				is_approved ? cw::fragf("%s\n", _("This action is currently approved; it will be selected whenever possible.")) :
+				cw::text_fragment(""),
+				// ForTranslators: the "r" here is
+				// for "reject", but should not be
+				// translated since the commands in
+				// aptitude's internal command-line
+				// aren't translated.
+				is_rejected ? cw::fragf(_("Enter \"r %s\" to allow this action to appear in new solutions."), tag.c_str())
+				// ForTranslators: the "r" here is
+				// for "reject", but should not be
+				// translated since the commands in
+				// aptitude's internal command-line
+				// aren't translated.
+                                            : cw::fragf(_("Enter \"r %s\" to prevent this action from appearing in new solutions."), tag.c_str()),
+				// ForTranslators: the "a" here is
+				// for "approve", but should not be
+				// translated since the commands in
+				// aptitude's internal command-line
+				// aren't translated.
+				is_approved ? cw::fragf(_("Enter \"a %s\" to cease requiring that new solutions include this action if possible."), tag.c_str())
+				// ForTranslators: the "a" here is
+				// for "approve", but should not be
+				// translated since the commands in
+				// aptitude's internal command-line
+				// aren't translated.
+                                            : cw::fragf(_("Enter \"a %s\" to require that new solutions include this action if possible."), tag.c_str()))));
+
+  const unsigned int screen_width = term->get_screen_width();
+  cwidget::fragment_contents lines = f->layout(screen_width, screen_width, cwidget::style());
+
+  delete f;
+
+  cout << lines << endl;
+}
+
+// The command-line blocks until the continuation hits, so we can just
+// run continuations in the background thread.
+void cmdline_resolver_trampoline(const sigc::slot<void> &f)
+{
+  f();
+}
+
 // TODO: make this generic?
 class cmdline_resolver_continuation : public resolver_manager::background_continuation
 {
@@ -467,9 +726,9 @@ public:
     abort();
   }
 
-  void aborted(const cwidget::util::Exception &e)
+  void aborted(const std::string &errmsg)
   {
-    retbox.put(resolver_result::Aborted(e.errmsg()));
+    retbox.put(resolver_result::Aborted(errmsg));
   }
 };
 
@@ -499,42 +758,11 @@ public:
   std::string errmsg() const { return msg; }
 };
 
-aptitude_solution calculate_current_solution(bool suppress_message)
+// Displays a spinner while we wait for the background thread to spit
+// out a solution.
+static aptitude_solution wait_for_solution(cwidget::threads::box<cmdline_resolver_continuation::resolver_result> &retbox,
+					   cmdline_spinner &spin)
 {
-  const int step_limit = aptcfg->FindI(PACKAGE "::ProblemResolver::StepLimit", 5000);
-  if(step_limit <= 0)
-    {
-      const std::string msg = ssprintf(_("Would resolve dependencies, but dependency resolution is disabled.\n   (%s::ProblemResolver::StepLimit = 0)\n"), PACKAGE);
-
-      // It's important that the code path leading to here doesn't
-      // access resman: the resolver won't exist in this case.
-      throw CmdlineSearchDisabledException(msg);
-    }
-
-  if(!resman->resolver_exists())
-    {
-      const std::string msg = _("I want to resolve dependencies, but no dependency resolver was created.");
-
-      throw CmdlineSearchAbortedException(msg);
-    }
-
-
-
-  if(resman->get_selected_solution() < resman->generated_solution_count())
-    return resman->get_solution(resman->get_selected_solution(), 0);
-
-
-  cmdline_spinner spin(aptcfg->FindI("Quiet", 0));
-
-  if(!suppress_message)
-    std::cout << _("Resolving dependencies...") << std::endl;
-
-  cwidget::threads::box<cmdline_resolver_continuation::resolver_result> retbox;
-
-  resman->get_solution_background(resman->generated_solution_count(),
-				  step_limit,
-				  new cmdline_resolver_continuation(retbox));
-
   cmdline_resolver_continuation::resolver_result res;
   bool done = false;
   // The number of milliseconds to step per display.
@@ -577,6 +805,47 @@ aptitude_solution calculate_current_solution(bool suppress_message)
     return res.sol;
 }
 
+aptitude_solution calculate_current_solution(bool suppress_message,
+                                             const shared_ptr<terminal> &term)
+{
+  const int step_limit = aptcfg->FindI(PACKAGE "::ProblemResolver::StepLimit", 5000);
+  if(step_limit <= 0)
+    {
+      const std::string msg = ssprintf(_("Would resolve dependencies, but dependency resolution is disabled.\n   (%s::ProblemResolver::StepLimit = 0)\n"), PACKAGE);
+
+      // It's important that the code path leading to here doesn't
+      // access resman: the resolver won't exist in this case.
+      throw CmdlineSearchDisabledException(msg);
+    }
+
+  if(!resman->resolver_exists())
+    {
+      const std::string msg = _("I want to resolve dependencies, but no dependency resolver was created.");
+
+      throw CmdlineSearchAbortedException(msg);
+    }
+
+
+
+  if(resman->get_selected_solution() < resman->generated_solution_count())
+    return resman->get_solution(resman->get_selected_solution(), 0);
+
+
+  cmdline_spinner spin(aptcfg->FindI("Quiet", 0), term);
+
+  if(!suppress_message)
+    std::cout << _("Resolving dependencies...") << std::endl;
+
+  cwidget::threads::box<cmdline_resolver_continuation::resolver_result> retbox;
+
+  resman->get_solution_background(resman->generated_solution_count(),
+				  step_limit,
+				  boost::make_shared<cmdline_resolver_continuation>(boost::ref(retbox)),
+				  cmdline_resolver_trampoline);
+
+  return wait_for_solution(retbox, spin);
+}
+
 aptitude::cmdline::cmdline_resolver_result
 cmdline_resolve_deps(pkgset &to_install,
 		     pkgset &to_hold,
@@ -586,7 +855,8 @@ cmdline_resolve_deps(pkgset &to_install,
 		     bool force_no_change,
 		     int verbose,
 		     pkgPolicy &policy,
-		     bool arch_only)
+		     bool arch_only,
+                     const shared_ptr<terminal> &term)
 {
   bool story_is_default = aptcfg->FindB(PACKAGE "::CmdLine::Resolver-Show-Steps", false);
 
@@ -596,31 +866,42 @@ cmdline_resolve_deps(pkgset &to_install,
 		     force_no_change);
       aptitude_solution lastsol;
 
+      // Stores the string IDs that can be used for accept/reject
+      // commands.  Filled in when the solution is being rendered
+      // (since the IDs are integers counting from the first entry in
+      // the displayed lists).
+      std::map<std::string, choice> ids;
+
+
       // The inner loop tries to generate solutions until some
       // packages are modified by the user (then the new set of broken
       // packages, if any, is displayed and we start over)
       bool modified_pkgs=false;
+
+      bool redisplay = false;
       while(!modified_pkgs)
 	try
 	  {
 	    try
 	      {
-		aptitude_solution sol = calculate_current_solution(true);
+		aptitude_solution sol = calculate_current_solution(true, term);
 
 		if(_error->PendingError())
 		  _error->DumpErrors();
 
-		if(sol != lastsol)
+		if(sol != lastsol || redisplay)
 		  {
+		    ids.clear();
+		    // \todo display "the following actions..." only
+		    // the first time through.
 		    cw::fragment *f=cw::sequence_fragment(flowbox(cwidget::text_fragment(_("The following actions will resolve these dependencies:"))),
 						  cwidget::newline_fragment(),
 						  story_is_default
-						    ? solution_story(sol)
-						    : solution_fragment(sol),
+						    ? solution_story(sol, &ids)
+						    : solution_fragment_with_ids(sol, ids),
 						  NULL);
 
-		    update_screen_width();
-
+                    const unsigned int screen_width = term->get_screen_width();
 		    cwidget::fragment_contents lines=f->layout(screen_width, screen_width, cwidget::style());
 
 		    delete f;
@@ -629,6 +910,8 @@ cmdline_resolve_deps(pkgset &to_install,
 		    lastsol=sol;
 		  }
 
+		redisplay = false;
+
 		string response=assume_yes?"Y":prompt_string(_("Accept this solution? [Y/n/q/?] "));
 
 		string::size_type loc=0;
@@ -643,7 +926,7 @@ cmdline_resolve_deps(pkgset &to_install,
 		switch(toupper(response[loc]))
 		  {
 		  case 'Y':
-		    (*apt_cache_file)->apply_solution(calculate_current_solution(true), NULL);
+		    (*apt_cache_file)->apply_solution(calculate_current_solution(true, term), NULL);
 		    modified_pkgs=true;
 		    break;
 		  case 'N':
@@ -664,10 +947,11 @@ cmdline_resolve_deps(pkgset &to_install,
 		  case 'O':
 		    {
 		      story_is_default = !story_is_default;
+		      ids.clear();
 		      cw::fragment *f = story_is_default
-			              ? solution_story(sol)
-			              : solution_fragment(sol);
-		      update_screen_width();
+			? solution_story(sol, &ids)
+			: solution_fragment_with_ids(sol, ids);
+                      const unsigned int screen_width = term->get_screen_width();
 		      cout << f->layout(screen_width, screen_width,
 					cwidget::style()) << endl;
 		      delete f;
@@ -677,10 +961,12 @@ cmdline_resolve_deps(pkgset &to_install,
 		    ui_solution_screen();
 		    break;
 		  case 'R':
-		    reject_or_mandate_version(string(response, 1), true);
+		    reject_or_mandate_version(string(response, 1), ids, true);
+		    redisplay = true;
 		    break;
 		  case 'A':
-		    reject_or_mandate_version(string(response, 1), false);
+		    reject_or_mandate_version(string(response, 1), ids, false);
+		    redisplay = true;
 		    break;
 		  case '.':
 		    resman->select_next_solution();
@@ -690,7 +976,7 @@ cmdline_resolve_deps(pkgset &to_install,
 		    break;
 		  case '?':
 		    cout << _("The following commands are available:") << endl;
-		    resolver_help(cout);
+		    resolver_help(cout, term);
 		    break;
 		  case '+':
 		  case '-':
@@ -702,7 +988,8 @@ cmdline_resolve_deps(pkgset &to_install,
 		      cmdline_parse_action(response, seen_virtual_packages,
 					   to_install, to_hold,
 					   to_remove, to_purge, verbose,
-					   policy, arch_only, false);
+					   policy, arch_only, false,
+                                           term);
 		      modified_pkgs=true;
 		    }
 		    break;
@@ -724,8 +1011,32 @@ cmdline_resolve_deps(pkgset &to_install,
 		    }
 		    break;
 		  default:
-		    cout << _("Invalid response; please enter one of the following commands:") << endl;
-		    resolver_help(cout);
+		    // Look for a choice tag in the first word of the
+		    // string.  NB: checking only the first word
+		    // requires special knowledge that choice tags are
+		    // in fact single words.
+		    {
+		      std::string first_word;
+
+		      while(loc < response.size() && !isspace(response[loc]))
+			{
+			  first_word += response[loc];
+			  ++loc;
+			}
+
+		      std::map<std::string, choice>::const_iterator found =
+			ids.find(first_word);
+		      if(found == ids.end())
+			{
+			  cout << _("Invalid response; please enter one of the following commands:") << endl;
+			  resolver_help(cout, term);
+			}
+		      else
+			{
+			  const choice &c = found->second;
+			  cmdline_resolver_show_choice(c, first_word, term);
+			}
+		    }
 		    break;
 		  }
 	      }
@@ -753,7 +1064,7 @@ cmdline_resolve_deps(pkgset &to_install,
 		      case 'Y':
 			try
 			  {
-			    calculate_current_solution(false);
+			    calculate_current_solution(false, term);
 			    done=true;
 			  }
 			catch(NoMoreTime)
@@ -827,178 +1138,75 @@ namespace aptitude
 {
   namespace cmdline
   {
-    namespace
+    // Implements the --show-resolver-actions command-line parameters.
+    void show_resolver_actions(const generic_solution<aptitude_universe> &solution,
+                               const shared_ptr<terminal> &term)
     {
-      // Set up aptitude's internal resolver for 'safe' resolution.
-      // Essentialy this means (1) forbid it from removing any package,
-      // and (2) only install the default candidate version of a package
-      // or the current version.
-      void setup_safe_resolver(bool no_new_installs, bool no_new_upgrades)
-      {
-	if(!resman->resolver_exists())
-	  return;
-
-	resman->reset_resolver();
-
-	resman->set_debug(aptcfg->FindB(PACKAGE "::CmdLine::Resolver-Debug", false));
-
-	for(pkgCache::PkgIterator p = (*apt_cache_file)->PkgBegin();
-	    !p.end(); ++p)
-	  {
-	    // Forbid the resolver from removing installed packages.
-	    if(!p.CurrentVer().end())
-	      {
-		aptitude_resolver_version
-		  remove_p(p,
-			   pkgCache::VerIterator(*apt_cache_file),
-			   *apt_cache_file);
-
-		resman->reject_version(remove_p);
-	      }
-
-	    // Forbid all real versions that aren't the current version or
-	    // the candidate version.
-	    for(pkgCache::VerIterator v = p.VersionList();
-		!v.end(); ++v)
-	      {
-		// For our purposes, all half-unpacked etc states are
-		// installed.
-		const bool p_is_installed =
-		  p->CurrentState != pkgCache::State::NotInstalled &&
-		  p->CurrentState != pkgCache::State::ConfigFiles;
-
-		const bool p_will_install =
-		  (*apt_cache_file)[p].Install();
-
-		if(v != p.CurrentVer() &&
-		   // Disallow installing not-installed packages that
-		   // aren't marked for installation if
-		   // no_new_installs is set.
-		   ((!p_is_installed && !p_will_install && no_new_installs) ||
-		    (p_is_installed && p_will_install && no_new_upgrades) ||
-		    v != (*apt_cache_file)[p].CandidateVerIter(*apt_cache_file)))
-		  {
-		    aptitude_resolver_version
-		      p_v(p, v, *apt_cache_file);
-		    resman->reject_version(p_v);
-		  }
-	      }
-	  }
-
-	cmdline_dump_resolver();
-      }
+      if(solution.get_choices().size() > 0)
+	{
+	  // The previous line will say "resolving dependencies...";
+	  // separate the solution from this message..
+	  std::cout << std::endl;
+	  std::auto_ptr<cw::fragment> story(solution_story(solution, NULL));
+          const unsigned int screen_width = term->get_screen_width();
+	  std::cout << story->layout(screen_width, screen_width, cwidget::style());
+	}
     }
 
     // Take the first solution we can compute, returning false if we
     // failed to find a solution.
-    bool safe_resolve_deps(int verbose, bool no_new_installs, bool no_new_upgrades)
+    bool safe_resolve_deps(int verbose,
+                           bool no_new_installs,
+                           bool no_new_upgrades,
+                           bool show_story,
+                           const shared_ptr<terminal> &term)
     {
       if(!resman->resolver_exists())
 	return true;
 
-      const bool debug = aptcfg->FindB(PACKAGE "::CmdLine::Resolver-Debug", false);
+      cmdline_dump_resolver();
 
-      setup_safe_resolver(no_new_installs, no_new_upgrades);
-
-      generic_solution<aptitude_universe> last_sol;
       try
 	{
-	  generic_solution<aptitude_universe> sol = calculate_current_solution(true);
-	  last_sol = sol;
-	  bool first = true;
-	  typedef imm::map<aptitude_resolver_package,
-	    generic_solution<aptitude_universe>::action> actions_map;
-	  while(!(!first &&
-		  last_sol &&
-		  sol.get_actions() == last_sol.get_actions() &&
-		  sol.get_unresolved_soft_deps() == last_sol.get_unresolved_soft_deps()))
-	    {
-	      first = false;
-	      // Mandate all the upgrades and installs in this solution,
-	      // then ask for the next solution.  The goal is to try to
-	      // find the best solution, not just some solution, but to
-	      // preserve our previous decisions in order to avoid
-	      // thrashing around between alternatives.
-	      for(actions_map::const_iterator it = sol.get_actions().begin();
-		  it != sol.get_actions().end(); ++it)
-		{
-		  if(it->second.ver.get_ver() != it->second.ver.get_pkg().CurrentVer())
-		    {
-		      if(debug)
-			std::cout << "Mandating the upgrade or install " << it->second.ver
-				  << ", since it was produced as part of a solution."
-				  << std::endl;
-		      resman->mandate_version(it->second.ver);
-		    }
-		  else
-		    {
-		      if(debug)
-			std::cout << "Not mandating " << it->second.ver
-				  << ", since it is a hold."
-				  << std::endl;
-		    }
-		}
-	      last_sol = sol;
-	      resman->select_next_solution();
-	      sol = calculate_current_solution(false);
-	    }
+	  cwidget::threads::box<cmdline_resolver_continuation::resolver_result> retbox;
 
-	  // Internal error that should never happen, but try to survive
-	  // if it does.
-	  std::cout << "*** Internal error: the resolver unexpectedly produced the same result twice."
-		    << std::endl
-		    << "Previous iteration was:" << std::endl;
-	  last_sol.dump(std::cout, true);
-	  std::cout << std::endl << "Current iteration was:" << std::endl;
-	  sol.dump(std::cout, true);
-	  std::cout << std::endl;
+	  resman->safe_resolve_deps_background(no_new_installs, no_new_upgrades,
+					       boost::make_shared<cmdline_resolver_continuation>(boost::ref(retbox)),
+					       cmdline_resolver_trampoline);
 
-	  (*apt_cache_file)->apply_solution(last_sol, NULL);
+	  cmdline_spinner spin(aptcfg->FindI("Quiet", 0), term);
+	  // TODO: maybe we should say "calculating upgrade" if we're
+	  // running safe-upgrade?
+	  std::cout << _("Resolving dependencies...") << std::endl;
+	  generic_solution<aptitude_universe> sol = wait_for_solution(retbox, spin);
+
+	  if(show_story)
+	    show_resolver_actions(sol, term);
+
+	  (*apt_cache_file)->apply_solution(sol, NULL);
 	}
       // If anything goes wrong, we give up (silently if verbosity is disabled).
       catch(NoMoreTime)
 	{
-	  if(last_sol)
-	    {
-	      std::cout << _("The resolver timed out after producing a solution; some possible upgrades might not be performed.");
-	      (*apt_cache_file)->apply_solution(last_sol, NULL);
-	      return true;
-	    }
-	  else
-	    {
-	      if(verbose > 0)
-		std::cout << _("Unable to resolve dependencies for the upgrade (the resolver timed out).");
-	      return false;
-	    }
+	  std::cout << cw::util::ssprintf(_("Unable to resolve dependencies for the upgrade because the resolver timed out.\n  You may be able to solve this problem by increasing\n  Aptitude::ProblemResolver::StepLimit (currently %d)."),
+					  // TODO: bad magic number
+					  // here -- get it from the
+					  // backend directly instead.
+					  aptcfg->FindI(PACKAGE "::ProblemResolver::StepLimit", 500000))
+		    << std::endl;
+	  return false;
 	}
       catch(NoMoreSolutions)
 	{
-	  if(last_sol)
-	    {
-	      (*apt_cache_file)->apply_solution(last_sol, NULL);
-	      return true;
-	    }
-	  else
-	    {
-	      if(verbose > 0)
-		std::cout << _("Unable to resolve dependencies for the upgrade (no solution found).");
-	      return false;
-	    }
+	  std::cout << _("Unable to resolve dependencies for the upgrade: no solution found.")
+		    << std::endl;
+	  return false;
 	}
       catch(const cw::util::Exception &e)
 	{
-	  if(last_sol)
-	    {
-	      (*apt_cache_file)->apply_solution(last_sol, NULL);
-	      std::cout << cw::util::ssprintf(_("Dependency resolution incomplete (%s); some possible upgrades might not be performed."), e.errmsg().c_str());
-	      return true;
-	    }
-	  else
-	    {
-	      if(verbose > 0)
-		std::cout << cw::util::ssprintf(_("Unable to resolve dependencies for the upgrade (%s)."), e.errmsg().c_str());
-	      return false;
-	    }
+	  std::cout << cw::util::ssprintf(_("Unable to resolve dependencies for the upgrade: %s"), e.errmsg().c_str())
+		    << std::endl;
+	  return false;
 	}
 
       return true;
