@@ -1,19 +1,42 @@
 // cmdline_show.cc                               -*-c++-*-
 //
-//  Copyright 2004 Daniel Burrows
+// Copyright (C) 2004, 2010 Daniel Burrows
+//
+// This program is free software; you can redistribute it and/or
+// modify it under the terms of the GNU General Public License as
+// published by the Free Software Foundation; either version 2 of the
+// License, or (at your option) any later version.
+//
+// This program is distributed in the hope that it will be useful, but
+// WITHOUT ANY WARRANTY; without even the implied warranty of
+// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+// General Public License for more details.
+//
+// You should have received a copy of the GNU General Public License
+// along with this program; see the file COPYING.  If not, write to
+// the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
+// Boston, MA 02111-1307, USA.
 
+
+// Local includes:
 #include "cmdline_show.h"
 
 #include <aptitude.h>
-#include <desc_parse.h>
+#include <desc_render.h>
 
 #include "cmdline_common.h"
 #include "cmdline_util.h"
+#include "terminal.h"
+#include "text_progress.h"
 
 #include <generic/apt/apt.h>
 #include <generic/apt/config_signal.h>
-#include <generic/apt/matchers.h>
+#include <generic/apt/matching/match.h>
+#include <generic/apt/matching/parse.h>
+#include <generic/apt/matching/pattern.h>
 
+
+// System includes:
 #include <cwidget/fragment.h>
 #include <cwidget/generic/util/transcode.h>
 
@@ -28,6 +51,12 @@
 #include <iostream>
 
 namespace cw = cwidget;
+using aptitude::cmdline::create_terminal;
+using aptitude::cmdline::create_terminal_locale;
+using aptitude::cmdline::make_text_progress;
+using aptitude::cmdline::terminal;
+using aptitude::cmdline::terminal_locale;
+using boost::shared_ptr;
 using cwidget::fragf;
 using cwidget::fragment;
 using namespace std;
@@ -230,14 +259,14 @@ static cwidget::fragment *deletion_fragment(const char *statestr,
       if(unused_delete)
 	return cw::fragf(_("%s; will be purged because nothing depends on it"), statestr);
       else
-	return cw::fragf(_("%s; will be purged"));
+	return cw::fragf(_("%s; will be purged"), statestr);
     }
   else
     {
       if(unused_delete)
 	return cw::fragf(_("%s; will be removed because nothing depends on it"), statestr);
       else
-	return cw::fragf(_("%s; will be removed"));
+	return cw::fragf(_("%s; will be removed"), statestr);
     }
 }
 
@@ -285,7 +314,7 @@ static cwidget::fragment *state_fragment(pkgCache::PkgIterator pkg, pkgCache::Ve
   if(!ver.end())
     {
       if(instver != ver && curver != ver)
-	return cw::fragf("%s", statestr);
+	return cw::fragf("%s%s", statestr, holdstr);
       else if(instver == ver)
 	{
 	  if(curver != instver)
@@ -294,15 +323,15 @@ static cwidget::fragment *state_fragment(pkgCache::PkgIterator pkg, pkgCache::Ve
 		 pkg->CurrentState == pkgCache::State::ConfigFiles)
 		{
 		  if(state.Flags & pkgCache::Flag::Auto)
-		    return cw::fragf(_("%s; will be installed"), statestr);
+		    return cw::fragf(_("%s%s; will be installed"), statestr, holdstr);
 		  else
-		    return cw::fragf(_("%s; will be installed automatically"), statestr);
+		    return cw::fragf(_("%s%s; will be installed automatically"), statestr, holdstr);
 		}
 	      else
 		return version_change_fragment(statestr, holdstr, curver, instver);
 	    }
 	  else
-	    return cw::fragf("%s", statestr);
+	    return cw::fragf("%s%s", statestr, holdstr);
 	}
       else if(instver.end()) // Now curver is ver.
 	return deletion_fragment(statestr, state, estate);
@@ -341,7 +370,8 @@ static cwidget::fragment *state_fragment(pkgCache::PkgIterator pkg, pkgCache::Ve
 }
 
 /** \brief Shows information about a package. */
-static void show_package(pkgCache::PkgIterator pkg, int verbose)
+static void show_package(pkgCache::PkgIterator pkg, int verbose,
+                         const shared_ptr<terminal> &term)
 {
   vector<cw::fragment *> fragments;
 
@@ -351,14 +381,15 @@ static void show_package(pkgCache::PkgIterator pkg, int verbose)
 
   cw::fragment *f=cw::sequence_fragment(fragments);
 
+  const unsigned int screen_width = term->get_screen_width();
   cout << f->layout(screen_width, screen_width, cwidget::style());
 
   delete f;
 }
 
-static cw::fragment *version_file_fragment(pkgCache::VerIterator ver,
-				       pkgCache::VerFileIterator vf,
-				       int verbose)
+cw::fragment *version_file_fragment(const pkgCache::VerIterator &ver,
+				    const pkgCache::VerFileIterator &vf,
+				    int verbose)
 {
   vector<cw::fragment *> fragments;
 
@@ -388,7 +419,7 @@ static cw::fragment *version_file_fragment(pkgCache::VerIterator ver,
 
   fragments.push_back(cw::fragf("%s%s%n", _("Version: "), ver.VerStr()));
   fragments.push_back(cw::fragf("%s%s%n", _("Priority: "),
-			    ver.PriorityType()?ver.PriorityType():_("N/A")));
+				const_cast<pkgCache::VerIterator &>(ver).PriorityType() ? const_cast<pkgCache::VerIterator &>(ver).PriorityType() : _("N/A")));
   fragments.push_back(cw::fragf("%s%s%n", _("Section: "),
 			    ver.Section()?ver.Section():_("N/A")));
   fragments.push_back(cw::fragf("%s%s%n", _("Maintainer: "),
@@ -453,12 +484,14 @@ static cw::fragment *version_file_fragment(pkgCache::VerIterator ver,
   return cw::sequence_fragment(fragments);
 }
 
-static void show_version(pkgCache::VerIterator ver, int verbose)
+static void show_version(pkgCache::VerIterator ver, int verbose,
+                         const shared_ptr<terminal> &term)
 {
   if(ver.FileList().end())
     {
       cw::fragment *f=version_file_fragment(ver, ver.FileList(), verbose);
 
+      const unsigned int screen_width = term->get_screen_width();
       cout << f->layout(screen_width, screen_width, cwidget::style());
 
       delete f;
@@ -469,6 +502,7 @@ static void show_version(pkgCache::VerIterator ver, int verbose)
 	{
 	  cw::fragment *f=version_file_fragment(ver, vf, verbose);
 
+          const unsigned int screen_width = term->get_screen_width();
 	  cout << f->layout(screen_width, screen_width, cwidget::style()) << endl;
 
 	  delete f;
@@ -485,7 +519,8 @@ bool do_cmdline_show_target(const pkgCache::PkgIterator &pkg,
 			    cmdline_version_source source,
 			    const string &sourcestr,
 			    int verbose,
-			    bool has_explicit_source)
+			    bool has_explicit_source,
+                            const shared_ptr<terminal> &term)
 {
   if(verbose == 0 || has_explicit_source)
     {
@@ -499,20 +534,20 @@ bool do_cmdline_show_target(const pkgCache::PkgIterator &pkg,
 	ver = pkg.VersionList();
 
       if(!ver.end())
-	show_version(ver, verbose);
+	show_version(ver, verbose, term);
       else
-	show_package(pkg, verbose);
+	show_package(pkg, verbose, term);
     }
   else if(!pkg.VersionList().end())
     for(pkgCache::VerIterator ver=pkg.VersionList(); !ver.end(); ++ver)
-      show_version(ver, verbose);
+      show_version(ver, verbose, term);
   else
-    show_package(pkg, verbose);
+    show_package(pkg, verbose, term);
 
   return true;
 }
 
-bool do_cmdline_show(string s, int verbose)
+bool do_cmdline_show(string s, int verbose, const shared_ptr<terminal> &term)
 {
   cmdline_version_source source;
   string name, sourcestr;
@@ -530,7 +565,7 @@ bool do_cmdline_show(string s, int verbose)
       sourcestr = default_release;
     }
 
-  bool is_pattern = cmdline_is_search_pattern(name);
+  bool is_pattern = aptitude::matching::is_pattern(name);
   pkgCache::PkgIterator pkg;
 
   if(!is_pattern)
@@ -545,26 +580,43 @@ bool do_cmdline_show(string s, int verbose)
     }
 
   if(!is_pattern && !pkg.end())
-    return do_cmdline_show_target(pkg, source, sourcestr, verbose, has_explicit_source);
+    return do_cmdline_show_target(pkg,
+                                  source,
+                                  sourcestr,
+                                  verbose,
+                                  has_explicit_source,
+                                  term);
   else if(is_pattern)
     {
       using namespace aptitude::matching;
+      using cwidget::util::ref_ptr;
 
-      std::auto_ptr<pkg_matcher> m(parse_pattern(name));
+      ref_ptr<pattern> p(parse(name));
 
-      if(m.get() == NULL)
+      if(!p.valid())
 	{
 	  _error->Error(_("Unable to parse pattern %s"), name.c_str());
 	  return false;
 	}
 
-      for(pkgCache::PkgIterator P=(*apt_cache_file)->PkgBegin();
-	  !P.end(); ++P)
-	if(apply_matcher(m.get(), P, *apt_cache_file, *apt_package_records))
-	  {
-	    if(!do_cmdline_show_target(P, source, sourcestr, verbose, has_explicit_source))
-	      return false;
-	  }
+      std::vector<std::pair<pkgCache::PkgIterator, ref_ptr<structural_match> > > matches;
+      ref_ptr<search_cache> search_info(search_cache::create());
+      search(p, search_info,
+	     matches,
+	     *apt_cache_file,
+	     *apt_package_records);
+
+      for(std::vector<std::pair<pkgCache::PkgIterator, ref_ptr<structural_match> > >::const_iterator
+	    it = matches.begin(); it != matches.end(); ++it)
+	{
+	  if(!do_cmdline_show_target(it->first,
+                                     source,
+                                     sourcestr,
+                                     verbose,
+                                     has_explicit_source,
+                                     term))
+	    return false;
+	}
     }
   else
     ; // TODO: print an error message -- Christian will kill me if I
@@ -575,10 +627,13 @@ bool do_cmdline_show(string s, int verbose)
 
 int cmdline_show(int argc, char *argv[], int verbose)
 {
+  shared_ptr<terminal> term = create_terminal();
+  shared_ptr<terminal_locale> term_locale = create_terminal_locale();
+
   _error->DumpErrors();
 
-  OpProgress progress;
-  apt_init(&progress, false);
+  shared_ptr<OpProgress> progress = make_text_progress(true, term, term_locale);
+  apt_init(progress.get(), false);
 
   if(_error->PendingError())
     {
@@ -587,7 +642,7 @@ int cmdline_show(int argc, char *argv[], int verbose)
     }
 
   for(int i=1; i<argc; ++i)
-    if(!do_cmdline_show(argv[i], verbose))
+    if(!do_cmdline_show(argv[i], verbose, term))
       {
 	_error->DumpErrors();
 	return -1;
