@@ -15,8 +15,7 @@
 #include <stdio.h>
 #include <string.h>
 
-/** Shows broken dependencies for a single package */
-static void show_broken_deps(pkgCache::PkgIterator pkg)
+void show_broken_deps(pkgCache::PkgIterator pkg)
 {
   unsigned int indent=strlen(pkg.Name())+3;
   bool is_first_dep=true;
@@ -80,7 +79,13 @@ static void show_broken_deps(pkgCache::PkgIterator pkg)
 			       target.CurrentVer().VerStr());
 		    }
 		  else
-		    printf(_("but it is not installable"));
+		    {
+		      pkgCache::VerIterator cand = (*apt_cache_file)[target].CandidateVerIter(*apt_cache_file);
+		      if(cand.end())
+			printf(_("but it is not installable."));
+		      else
+			printf(_("but it is not going to be installed."));
+		    }
 		}
 	      else
 		// FIXME: do something sensible here!
