@@ -1,17 +1,45 @@
 // cmdline_simulate.h               -*-c++-*-
 //
-//  Copyright 2004 Daniel Burrows
+// Copyright (C) 2004, 2010 Daniel Burrows
+//
+// This program is free software; you can redistribute it and/or
+// modify it under the terms of the GNU General Public License as
+// published by the Free Software Foundation; either version 2 of the
+// License, or (at your option) any later version.
+//
+// This program is distributed in the hope that it will be useful, but
+// WITHOUT ANY WARRANTY; without even the implied warranty of
+// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+// General Public License for more details.
+//
+// You should have received a copy of the GNU General Public License
+// along with this program; see the file COPYING.  If not, write to
+// the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
+// Boston, MA 02111-1307, USA.
 
 #ifndef CMDLINE_SIMULATE_H
 #define CMDLINE_SIMULATE_H
 
+// Local includes:
 #include "cmdline_common.h"
 
+
+// System includes:
+#include <boost/shared_ptr.hpp>
+
 /** \file cmdline_simulate.h
  */
 
 class pkgPolicy;
 
+namespace aptitude
+{
+  namespace cmdline
+  {
+    class terminal;
+  }
+}
+
 /** Simulate an install run.  to_install and friends are meant to be
  *  the sets the user explicitly selected (so the prompt can be
  *  displayed only when extra stuff is added or removed).
@@ -44,7 +72,8 @@ int cmdline_simulate(bool as_upgrade,
 		     bool showsize, bool showwhy,
 		     bool always_prompt, int verbose,
 		     bool assume_yes, bool force_no_change,
-		     pkgPolicy &policy, bool arch_only);
+		     pkgPolicy &policy, bool arch_only,
+                     const boost::shared_ptr<aptitude::cmdline::terminal> &term);
 
 
 #endif // CMDLINE_SIMULATE_H
