@@ -5,6 +5,9 @@
 #ifndef CMDLINE_UPDATE_H
 #define CMDLINE_UPDATE_H
 
-int cmdline_update(int argc, char *argv[]);
+/** \file cmdline_update.h
+ */
+
+int cmdline_update(int argc, char *argv[], int verbose);
 
 #endif // CMDLINE_UPDATE_H
