@@ -1,6 +1,6 @@
 // cmdline_user_tag.cc
 //
-//   Copyright (C) 2008 Daniel Burrows
+//   Copyright (C) 2008-2010 Daniel Burrows
 //
 //   This program is free software; you can redistribute it and/or
 //   modify it under the terms of the GNU General Public License as
@@ -17,9 +17,13 @@
 //   the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 //   Boston, MA 02111-1307, USA.
 
+
+// Local includes:
 #include "cmdline_user_tag.h"
 
 #include "cmdline_util.h"
+#include "terminal.h"
+#include "text_progress.h"
 
 #include <aptitude.h>
 
@@ -27,10 +31,20 @@
 
 #include <generic/apt/apt.h>
 #include <generic/apt/aptcache.h>
-#include <generic/apt/matchers.h>
+#include <generic/apt/matching/match.h>
+#include <generic/apt/matching/parse.h>
+#include <generic/apt/matching/pattern.h>
 
+#include <stdio.h>
 #include <string.h>
 
+using aptitude::cmdline::create_terminal;
+using aptitude::cmdline::create_terminal_locale;
+using aptitude::cmdline::make_text_progress;
+using aptitude::cmdline::terminal;
+using aptitude::cmdline::terminal_locale;
+using boost::shared_ptr;
+
 namespace aptitude
 {
   namespace cmdline
@@ -70,6 +84,9 @@ namespace aptitude
 
     int cmdline_user_tag(int argc, char *argv[], int quiet, int verbose)
     {
+      const shared_ptr<terminal> term = create_terminal();
+      const shared_ptr<terminal_locale> term_locale = create_terminal_locale();
+
       user_tag_action action = (user_tag_action)-1;
 
       if(strcmp(argv[0], "add-user-tag") == 0)
@@ -107,7 +124,7 @@ namespace aptitude
       bool all_ok = true;
       for(int i = 2; i < argc; ++i)
 	{
-	  if(!cmdline_is_search_pattern(argv[i]))
+	  if(!aptitude::matching::is_pattern(argv[i]))
 	    {
 	      pkgCache::PkgIterator pkg = (*apt_cache_file)->FindPkg(argv[i]);
 	      if(pkg.end())
@@ -121,26 +138,34 @@ namespace aptitude
 	    }
 	  else
 	    {
-	      matching::pkg_matcher *m = matching::parse_pattern(argv[i]);
-	      if(m == NULL)
+	      using namespace aptitude::matching;
+	      using cwidget::util::ref_ptr;
+
+	      ref_ptr<pattern> p(parse(argv[i]));
+
+	      if(!p.valid())
 		{
 		  _error->DumpErrors();
 		  all_ok = false;
 		}
 	      else
 		{
-		  for(pkgCache::PkgIterator pkg = (*apt_cache_file)->PkgBegin();
-		      !pkg.end(); ++pkg)
-		    {
-		      if(matching::apply_matcher(m, pkg, *apt_cache_file, *apt_package_records))
-			do_user_tag(action, tag, pkg, verbose);
-		    }
+		  std::vector<std::pair<pkgCache::PkgIterator, ref_ptr<structural_match> > > matches;
+		  ref_ptr<search_cache> search_info(search_cache::create());
+		  search(p, search_info,
+			 matches,
+			 *apt_cache_file,
+			 *apt_package_records);
+
+		  for(std::vector<std::pair<pkgCache::PkgIterator, ref_ptr<structural_match> > >::const_iterator
+			it = matches.begin(); it != matches.end(); ++it)
+		    do_user_tag(action, tag, it->first, verbose);
 		}
 	    }
 	}
 
-      OpTextProgress text_progress;
-      if(!(*apt_cache_file)->save_selection_list(text_progress))
+      shared_ptr<OpProgress> text_progress = make_text_progress(false, term, term_locale);
+      if(!(*apt_cache_file)->save_selection_list(*text_progress))
 	return 1;
 
       if(!all_ok)
