@@ -1,12 +1,31 @@
 // cmdline_util.cc
 //
-//  Copyright 2004 Daniel Burrows
+// Copyright (C) 2004, 2010 Daniel Burrows
+//
+// This program is free software; you can redistribute it and/or
+// modify it under the terms of the GNU General Public License as
+// published by the Free Software Foundation; either version 2 of the
+// License, or (at your option) any later version.
+//
+// This program is distributed in the hope that it will be useful, but
+// WITHOUT ANY WARRANTY; without even the implied warranty of
+// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+// General Public License for more details.
+//
+// You should have received a copy of the GNU General Public License
+// along with this program; see the file COPYING.  If not, write to
+// the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
+// Boston, MA 02111-1307, USA.
 
+
+// Local includes:
 #include "cmdline_util.h"
 
 #include "cmdline_common.h"
 #include "cmdline_progress.h"
 #include "cmdline_show.h" // For operator<<
+#include "text_progress.h"
+#include "terminal.h"
 
 #include <aptitude.h>
 #include <ui.h>
@@ -16,7 +35,18 @@
 #include <generic/apt/config_signal.h>
 #include <generic/apt/download_manager.h>
 #include <generic/apt/download_signal_log.h>
-#include <generic/apt/matchers.h>
+#include <generic/apt/matching/match.h>
+#include <generic/apt/matching/parse.h>
+#include <generic/apt/matching/pattern.h>
+
+
+// System includes:
+#include <apt-pkg/error.h>
+#include <apt-pkg/indexfile.h>
+#include <apt-pkg/metaindex.h>
+#include <apt-pkg/pkgsystem.h>
+#include <apt-pkg/sourcelist.h>
+#include <apt-pkg/version.h>
 
 #include <cwidget/fragment.h>
 #include <cwidget/toplevel.h>
@@ -25,17 +55,15 @@
 #include <sigc++/bind.h>
 #include <sigc++/functors/ptr_fun.h>
 
-#include <apt-pkg/error.h>
-#include <apt-pkg/indexfile.h>
-#include <apt-pkg/metaindex.h>
-#include <apt-pkg/pkgsystem.h>
-#include <apt-pkg/sourcelist.h>
-#include <apt-pkg/version.h>
-
 #include <algorithm>
 
 namespace cw = cwidget;
 
+using aptitude::cmdline::make_text_progress;
+using aptitude::cmdline::terminal;
+using aptitude::cmdline::terminal_locale;
+using boost::shared_ptr;
+
 namespace
 {
   const int no_install_run_from_ui_preview_return = -1;
@@ -73,22 +101,25 @@ void ui_solution_screen()
   file_quit.connect(sigc::ptr_fun(cwidget::toplevel::exitmain));
 
   progress_ref p = gen_progress_bar();
-  do_new_package_view(*p.unsafe_get_ref());
+  do_new_package_view(*p->get_progress().unsafe_get_ref());
 
   do_examine_solution();
   ui_main();
   exit(0);
 }
 
-void cmdline_show_stringlist(strvector &items)
+void cmdline_show_stringlist(strvector &items,
+                             const shared_ptr<terminal> &term)
 {
+  const unsigned int screen_width = term->get_screen_width();
+
   int loc=2;
 
   printf("  ");
 
   for(strvector::iterator i=items.begin(); i!=items.end(); ++i)
     {
-      if(loc+i->size()>screen_width-5)
+      if(loc + i->size() > (unsigned)(screen_width - 5))
 	{
 	  printf("\n  ");
 	  loc=2;
@@ -101,14 +132,15 @@ void cmdline_show_stringlist(strvector &items)
   printf("\n");
 }
 
-void cmdline_show_pkglist(pkgvector &items)
+void cmdline_show_pkglist(pkgvector &items,
+                          const shared_ptr<terminal> &term)
 {
   strvector tmp;
 
   for(pkgvector::iterator i=items.begin(); i!=items.end(); ++i)
     tmp.push_back(i->Name());
 
-  cmdline_show_stringlist(tmp);
+  cmdline_show_stringlist(tmp, term);
 }
 
 pkgCache::VerIterator cmdline_find_ver(pkgCache::PkgIterator pkg,
@@ -145,6 +177,15 @@ pkgCache::VerIterator cmdline_find_ver(pkgCache::PkgIterator pkg,
 	      return ver;
 	  }
 
+      for(pkgCache::VerIterator ver = pkg.VersionList(); !ver.end(); ++ver)
+	for(pkgCache::VerFileIterator verfile = ver.FileList();
+	    !verfile.end(); ++verfile)
+	  {
+	    pkgCache::PkgFileIterator pkgfile = verfile.File();
+	    if(pkgfile.Codename() && sourcestr == pkgfile.Codename())
+	      return ver;
+	  }
+
       printf(_("Unable to find an archive \"%s\" for the package \"%s\"\n"),
 	     sourcestr.c_str(),
 	     pkg.Name());
@@ -277,7 +318,8 @@ namespace
 
   void show_stats_change(stats initial, stats final,
 			 bool show_all,
-			 bool show_unchanged)
+			 bool show_unchanged,
+                         const shared_ptr<terminal> &term)
   {
     using cw::fragf;
     using cw::util::ssprintf;
@@ -336,13 +378,13 @@ namespace
     std::set_difference(final.get_obsolete().begin(), final.get_obsolete().end(),
 			initial.get_obsolete().begin(), initial.get_obsolete().end(),
 			std::back_insert_iterator<std::vector<std::string> >(new_obsolete));
-    const int obsolete_list_limit = aptcfg->FindI(PACKAGE "::Max-Obsolete-List-Length", 50);
+    const unsigned int obsolete_list_limit = aptcfg->FindI(PACKAGE "::Max-Obsolete-List-Length", 50);
     if(new_obsolete.size() > obsolete_list_limit)
       {
 	output_fragments.push_back(cw::text_fragment(ssprintf(ngettext("There is %d newly obsolete package.",
 								       "There are %d newly obsolete packages.",
 								       new_obsolete.size()),
-							      new_obsolete.size())));
+							      (int)new_obsolete.size())));
       }
     else if(new_obsolete.size() > 0)
       {
@@ -362,18 +404,29 @@ namespace
       {
 	cw::fragment *f = join_fragments(output_fragments, L"\n");
 
-	update_screen_width();
+        const unsigned int screen_width = term->get_screen_width();
 	std::cout << f->layout(screen_width, screen_width, cw::style());
 	delete f;
       }
   }
 }
 
+namespace
+{
+  template<typename T>
+  void assign(const T &val, T *target)
+  {
+    *target = val;
+  }
+}
+
 download_manager::result cmdline_do_download(download_manager *m,
-					     int verbose)
+					     int verbose,
+                                             const shared_ptr<terminal> &term,
+                                             const shared_ptr<terminal_locale> &term_locale)
 {
   stats initial_stats(0, 0, 0, std::set<std::string>());
-  OpTextProgress progress(aptcfg->FindI("Quiet", 0));
+  shared_ptr<OpProgress> progress = make_text_progress(false, term, term_locale);
 
   if(aptcfg->FindI("Quiet", 0) == 0)
     {
@@ -387,13 +440,13 @@ download_manager::result cmdline_do_download(download_manager *m,
       initial_stats = compute_apt_stats();
     }
 
-  std::auto_ptr<download_signal_log> log(gen_cmdline_download_progress());
+  std::auto_ptr<download_signal_log> log(gen_cmdline_download_progress(term));
 
   // Dump errors here because prepare() might check for pending errors
   // and think something failed.
   _error->DumpErrors();
 
-  if(!m->prepare(progress, *log.get(), log.get()))
+  if(!m->prepare(*progress, *log.get(), log.get()))
     return download_manager::failure;
 
   download_manager::result finish_res;
@@ -401,7 +454,9 @@ download_manager::result cmdline_do_download(download_manager *m,
   do
     {
       pkgAcquire::RunResult download_res = m->do_download();
-      finish_res = m->finish(download_res, progress);
+      m->finish(download_res, progress.get(),
+		sigc::bind(sigc::ptr_fun(&assign<download_manager::result>),
+			   &finish_res));
     }
   while(finish_res == download_manager::do_again);
 
@@ -412,17 +467,13 @@ download_manager::result cmdline_do_download(download_manager *m,
       apt_load_cache(&tmpProgress, false, NULL);
       final_stats = compute_apt_stats();
       show_stats_change(initial_stats, final_stats,
-			verbose >= 1, verbose >= 2);
+			verbose >= 1, verbose >= 2,
+                        term);
     }
 
   return finish_res;
 }
 
-bool cmdline_is_search_pattern(const std::string &s)
-{
-  return s.find_first_of("~?") != s.npos;
-}
-
 namespace aptitude
 {
   namespace cmdline
@@ -559,6 +610,11 @@ namespace aptitude
 	  rval = find_source_package(source_package_name);
 	  break;
 
+	case cmdline_version_curr_or_cand:
+	  break;
+	  // Do this later since we need to refer to the installed
+	  // version.
+
 	case cmdline_version_archive:
 	  _error->DumpErrors();
 	  rval = find_source_by_archive(source_package_name, version_source_string);
@@ -593,6 +649,13 @@ namespace aptitude
 		  const std::string source_version =
 		    rec.SourceVer().empty() ? ver.VerStr() : rec.SourceVer();
 
+		  rval = find_source_package(source_package_name, source_version);
+		}
+	      else if(version_source == cmdline_version_curr_or_cand)
+		{
+		  const std::string source_version =
+		    rec.SourceVer().empty() ? ver.VerStr() : rec.SourceVer();		  
+
 		  rval = find_source_package(source_package_name, source_version);
 		}
 	    }
@@ -616,6 +679,9 @@ namespace aptitude
 	      rval = find_source_package(source_package_name);
 	      break;
 
+	    case cmdline_version_curr_or_cand:
+	      break; // We would have tried already if it was possible.
+
 	    case cmdline_version_archive:
 	      _error->DumpErrors();
 	      rval = find_source_by_archive(source_package_name, version_source_string);
@@ -633,6 +699,7 @@ namespace aptitude
     void apply_user_tags(const std::vector<tag_application> &user_tags)
     {
       using namespace matching;
+      cwidget::util::ref_ptr<search_cache> search_info(search_cache::create());
       for(pkgCache::PkgIterator pkg = (*apt_cache_file)->PkgBegin();
 	  !pkg.end(); ++pkg)
 	{
@@ -640,11 +707,13 @@ namespace aptitude
 		user_tags.begin(); it != user_tags.end(); ++it)
 	    {
 	      bool applicable = false;
-	      if(it->get_matcher() != NULL)
+	      if(it->get_pattern().valid())
 		{
-		  if(matching::apply_matcher(it->get_matcher(), pkg,
-				   *apt_cache_file,
-				   *apt_package_records))
+		  if(matching::get_match(it->get_pattern(),
+					 pkg,
+					 search_info,
+					 *apt_cache_file,
+					 *apt_package_records).valid())
 		    applicable = true;
 		}
 	      else
@@ -669,5 +738,44 @@ namespace aptitude
 	    }
 	}
     }
+
+    std::wstring de_columnize(const cwidget::config::column_definition_list &columns,
+			      cwidget::config::column_generator &columnizer,
+			      cwidget::config::column_parameters &p)
+    {
+      using namespace cwidget::config;
+
+      // TODO: this should move into cwidget in the future, as a new
+      // mode of operation for layout_columns().
+      std::wstring output;
+
+      for(column_definition_list::const_iterator it = columns.begin();
+	  it != columns.end();
+	  ++it)
+	{
+	  if(it->type == column_definition::COLUMN_LITERAL)
+	    output += it->arg;
+	  else
+	    {
+	      eassert(it->type == column_definition::COLUMN_GENERATED ||
+		      it->type == column_definition::COLUMN_PARAM);
+
+	      if(it->type == column_definition::COLUMN_GENERATED)
+		{
+		  cwidget::column_disposition disp = columnizer.setup_column(it->ival);
+		  output += disp.text;
+		}
+	      else
+		{
+		  if(p.param_count() <= it->ival)
+		    output += L"###";
+		  else
+		    output += p.get_param(it->ival);
+		}
+	    }
+	}
+
+      return output;
+    }
   }
 }
