@@ -1,6 +1,6 @@
 // apt.cc
 //
-//  Copyright 1999-2008 Daniel Burrows
+//  Copyright 1999-2010 Daniel Burrows
 //
 //  This program is free software; you can redistribute it and/or modify
 //  it under the terms of the GNU General Public License as published by
@@ -22,9 +22,11 @@
 #include "apt.h"
 
 #include <aptitude.h>
+#include <loggers.h>
 
-
+#include "aptitude_resolver_universe.h"
 #include "config_signal.h"
+#include "download_queue.h"
 #include "pkg_hier.h"
 #include "resolver_manager.h"
 #include "rev_dep_iterator.h"
@@ -34,6 +36,7 @@
 #include <cwidget/generic/util/eassert.h>
 #include <cwidget/generic/util/transcode.h>
 
+#include <generic/util/file_cache.h>
 #include <generic/util/util.h>
 
 #include <generic/util/undo.h>
@@ -54,6 +57,9 @@
 #include <errno.h>
 
 using namespace std;
+using aptitude::Loggers;
+
+namespace cw = cwidget;
 
 enum interesting_state {uncached = 0, uninteresting, interesting};
 static interesting_state *cached_deps_interesting = NULL;
@@ -62,17 +68,12 @@ static interesting_state *cached_deps_interesting = NULL;
 // pointer in the following table is set to 1 when a result is cached:
 static pkgCache::Dependency **cached_surrounding_or = NULL;
 
-aptitudeCacheFile *apt_cache_file=NULL;
-signalling_config *aptcfg=NULL;
-pkgRecords *apt_package_records=NULL;
-pkgSourceList *apt_source_list=NULL;
-undo_list *apt_undos=NULL;
 pkg_hier *user_pkg_hier=NULL;
-resolver_manager *resman = NULL;
 
 string *pendingerr=NULL;
 bool erroriswarning=false;
 
+
 // Set to "true" if we have a version of the apt library with
 // support for overriding configuration settings via RootDir.
 static bool apt_knows_about_rootdir;
@@ -119,8 +120,10 @@ bool get_apt_knows_about_rootdir()
   return apt_knows_about_rootdir;
 }
 
-void apt_preinit()
+void apt_preinit(const char *rootdir)
 {
+  logging::LoggerPtr logger(Loggers::getAptitudeAptGlobals());
+
   // The old name for the recommends-should-be-automatically-installed
   // setting and the new one.
   const char * const aptitudeIgnoreRecommendsImportant = PACKAGE "::Ignore-Recommends-Important";
@@ -154,6 +157,13 @@ void apt_preinit()
   theme_config=new Configuration;
   user_config=new Configuration;
 
+  if(rootdir != NULL)
+    {
+      _config->Set("RootDir", rootdir);
+      theme_config->Set("RootDir", rootdir);
+      user_config->Set("RootDir", rootdir);
+    }
+
   ReadConfigFile(*theme_config, PKGDATADIR "/aptitude-defaults");
 
   pkgInitConfig(*_config);
@@ -276,8 +286,15 @@ void apt_dumpcfg(const char *root)
       return;
     }
 
+  // Don't write RootDir to the user's configuration file -- it causes
+  // horrible confusion.
+  std::string rootDir(_config->Find("RootDir", ""));
+  _config->Clear("RootDir");
+
   aptcfg->Dump(f);
 
+  _config->Set("RootDir", rootDir);
+
   f.close();
 
   if(rename((cfgloc+".new").c_str(), cfgloc.c_str())!=0)
@@ -293,10 +310,16 @@ void apt_revertoptions()
   Configuration *old_user_config=user_config;
   Configuration *old_config=_config;
 
+  // Preserve any existing root-dir settings in the new configuration.
+  const std::string old_rootdir = _config->Find("RootDir", "");
+
   _config=new Configuration;
 
   user_config=new Configuration;
 
+  _config->Set("RootDir", old_rootdir);
+  user_config->Set("RootDir", old_rootdir);
+
   // ick?
   pkgInitConfig(*_config);
   pkgInitSystem(*_config, _system);
@@ -316,8 +339,14 @@ void apt_init(OpProgress *progress_bar, bool do_initselections,
 
 void apt_close_cache()
 {
+  logging::LoggerPtr logger(Loggers::getAptitudeAptGlobals());
+
+  LOG_INFO(logger, "Closing apt cache.");
+
   cache_closed();
 
+  LOG_TRACE(logger, "Done emitting cache_closed().");
+
   //   DANGER WILL ROBINSON!
   //
   // This must be done *** BEFORE BEFORE BEFORE *** we delete the
@@ -328,45 +357,76 @@ void apt_close_cache()
     {
       delete resman;
       resman = NULL;
+
+      LOG_TRACE(logger, "Deleted the global dependency resolver manager.");
     }
+  else
+    LOG_TRACE(logger, "No global dependency resolver manager exists; none deleted.");
 
   reset_tasks();
 
+  LOG_TRACE(logger, "Tasks reset.");
+
   if(apt_package_records)
     {
       delete apt_package_records;
       apt_package_records=NULL;
+      LOG_TRACE(logger, "Deleted the apt package records.");
     }
+  else
+    LOG_TRACE(logger, "No global apt package records exist; none deleted.");
 
   if(apt_cache_file)
     {
       delete apt_cache_file;
       apt_cache_file=NULL;
+      LOG_TRACE(logger, "Deleted the apt cache file.");
     }
+  else
+    LOG_TRACE(logger, "No global apt cache file exists; none deleted.");
 
   if(apt_source_list)
     {
       delete apt_source_list;
       apt_source_list=NULL;
+      LOG_TRACE(logger, "Deleted the apt sources list.");
     }
+  else
+    LOG_TRACE(logger, "No global apt sources list exists; none deleted.");
+
+  LOG_DEBUG(logger, "Done closing the apt cache.");
 }
 
 void apt_load_cache(OpProgress *progress_bar, bool do_initselections,
 		    const char * status_fname)
 {
+  logging::LoggerPtr logger(Loggers::getAptitudeAptGlobals());
+
   if(apt_cache_file != NULL)
-    return;
+    {
+      LOG_TRACE(logger, "Not loading apt cache: it's already loaded.");
+      return;
+    }
+
+  LOG_INFO(logger, "Loading apt cache.");
 
   aptitudeCacheFile *new_file=new aptitudeCacheFile;
 
+  LOG_TRACE(logger, "Reading the sources list.");
+
   apt_source_list=new pkgSourceList;
   apt_source_list->ReadMainList();
 
   bool simulate = aptcfg->FindB(PACKAGE "::Simulate", false);
 
+  if(simulate)
+    LOG_DEBUG(logger, PACKAGE "::Simulate is set; not locking the cache file.");
+
   // Clear the error stack so that we don't get confused by old errors.
   consume_errors();
 
+  LOG_TRACE(logger, "Opening the apt cache.");
+
   bool open_failed=!new_file->Open(*progress_bar, do_initselections,
 				   (getuid() == 0) && !simulate,
 				   status_fname)
@@ -374,12 +434,21 @@ void apt_load_cache(OpProgress *progress_bar, bool do_initselections,
 
   if(open_failed && getuid() == 0)
     {
+      // Hm, we should include the errors, but there's no
+      // nondestructive way to do that. :-(
+      LOG_ERROR(logger, "Failed to load the apt cache; trying to open it without locking.");
+
       // Don't discard errors, make sure they get displayed instead.
       consume_errors();
 
       open_failed=!new_file->Open(*progress_bar, do_initselections,
 				  false, status_fname);
 
+      if(open_failed)
+	LOG_ERROR(logger, "Unable to load the apt cache at all; giving up.");
+      else
+	LOG_DEBUG(logger, "Opening the apt cache with locking succeeded.");
+
       if(!open_failed)
 	_error->Warning(_("Could not lock the cache file; this usually means that dpkg or another apt tool is already installing packages.  Opening in read-only mode; any changes you make to the states of packages will NOT be preserved!"));
     }
@@ -387,7 +456,9 @@ void apt_load_cache(OpProgress *progress_bar, bool do_initselections,
   if(open_failed)
     {
       delete new_file;
+      LOG_DEBUG(logger, "Unable to load the apt cache; aborting and emitting cache_reload_failed().");
       cache_reload_failed();
+      LOG_TRACE(logger, "Done emitting cache_reload_failed().");
       return;
     }
 
@@ -398,14 +469,20 @@ void apt_load_cache(OpProgress *progress_bar, bool do_initselections,
   // ^C and lose important changes (like the new dselect states of
   // packages).  Note, though, that we don't fail if this fails.
   if(!status_fname && apt_cache_file->is_locked())
-    (*apt_cache_file)->save_selection_list(*progress_bar);
+    {
+      LOG_TRACE(logger, "Trying to save the current selection list.");
+      (*apt_cache_file)->save_selection_list(*progress_bar);
+    }
 
+  LOG_TRACE(logger, "Loading the apt package records.");
   apt_package_records=new pkgRecords(*apt_cache_file);
 
   // Um, good time to clear our undo info.
   apt_undos->clear_items();
 
+  LOG_TRACE(logger, "Loading task information.");
   load_tasks(*progress_bar);
+  LOG_TRACE(logger, "Loading tags.");
 #ifndef HAVE_EPT
   load_tags(*progress_bar);
 #else
@@ -414,13 +491,52 @@ void apt_load_cache(OpProgress *progress_bar, bool do_initselections,
 
   if(user_pkg_hier)
     {
+      LOG_TRACE(logger, "Loading user-defined package hierarchy information.");
       reload_user_pkg_hier();
       hier_reloaded();
     }
 
-  resman = new resolver_manager(*new_file);
+  LOG_TRACE(logger, "Initializing global dependency resolver manager.");
+  resman = new resolver_manager(new_file, imm::map<aptitude_resolver_package, aptitude_resolver_version>());
 
+  LOG_TRACE(logger, "Initializing the download cache.");
+  // Open the download cache.  By default, it goes in
+  // ~/.aptitude/cache; it has 512Kb of in-memory cache and 10MB of
+  // on-disk cache.
+  const char *HOME = getenv("HOME");
+  if(HOME != NULL)
+    {
+      std::string download_cache_file_name = string(HOME) + "/.aptitude/cache";
+      const int download_cache_memory_size =
+	aptcfg->FindI(PACKAGE "::UI::DownloadCache::MemorySize", 512 * 1024);
+      const int download_cache_disk_size   =
+	aptcfg->FindI(PACKAGE "::UI::DownloadCache::DiskSize", 10 * 1024 * 1024);
+      try
+	{
+	  download_cache = aptitude::util::file_cache::create(download_cache_file_name,
+							      download_cache_memory_size,
+							      download_cache_disk_size);
+	}
+      catch(cwidget::util::Exception &ex)
+	{
+	  LOG_WARN(logger,
+		   "Can't open the file cache \""
+		   << download_cache_file_name
+		   << "\": " << ex.errmsg());
+	}
+      catch(std::exception &ex)
+	{
+	  LOG_WARN(logger,
+		   "Can't open the file cache \""
+		   << download_cache_file_name
+		   << "\": " << ex.what());
+	}
+    }
+
+  LOG_DEBUG(logger, "Emitting cache_reloaded().");
   cache_reloaded();
+
+  LOG_TRACE(logger, "Done emitting cache_reloaded().");
 }
 
 void apt_reload_cache(OpProgress *progress_bar, bool do_initselections,
@@ -430,6 +546,41 @@ void apt_reload_cache(OpProgress *progress_bar, bool do_initselections,
   apt_load_cache(progress_bar, do_initselections, status_fname);
 }
 
+void apt_shutdown()
+{
+  aptitude::shutdown_download_queue();
+
+
+  apt_close_cache();
+
+  delete aptcfg;
+  aptcfg = NULL;
+
+  delete theme_config;
+  theme_config = NULL;
+
+  delete user_config;
+  user_config = NULL;
+
+  delete apt_undos;
+  apt_undos = NULL;
+
+  delete user_pkg_hier;
+  user_pkg_hier = NULL;
+
+  delete pendingerr;
+  pendingerr = NULL;
+
+
+  download_cache.reset();
+
+  cache_closed.clear();
+  cache_reloaded.clear();
+  cache_reload_failed.clear();
+  hier_reloaded.clear();
+  consume_errors.clear();
+}
+
 pkg_hier *get_user_pkg_hier()
 {
   if(!user_pkg_hier)
@@ -439,12 +590,13 @@ pkg_hier *get_user_pkg_hier()
 }
 
 pkg_action_state find_pkg_state(pkgCache::PkgIterator pkg,
-				aptitudeDepCache &cache)
+				aptitudeDepCache &cache,
+                                bool ignore_broken)
 {
   aptitudeDepCache::StateCache &state = cache[pkg];
   aptitudeDepCache::aptitude_state &extstate = cache.get_ext_state(pkg);
 
-  if(state.InstBroken())
+  if(state.InstBroken() && !ignore_broken)
     return pkg_broken;
   else if(state.Delete())
     {
@@ -711,6 +863,106 @@ bool package_trusted(const pkgCache::VerIterator &ver)
   return false;
 }
 
+pkgCache::VerIterator install_version(const pkgCache::PkgIterator &pkg,
+				      aptitudeDepCache &cache)
+{
+  if(pkg.VersionList().end())
+    return pkgCache::VerIterator(cache);
+
+  pkgDepCache::StateCache &state(cache[pkg]);
+
+  if(state.Delete())
+    return pkgCache::VerIterator(cache);
+  else if(state.Install())
+    return state.InstVerIter(cache);
+  else // state.Keep()
+    {
+      if(pkg->CurrentState == pkgCache::State::NotInstalled ||
+	 pkg->CurrentState == pkgCache::State::ConfigFiles)
+	return pkgCache::VerIterator(cache);
+      else
+	return pkg.CurrentVer();
+    }
+}
+
+pkgCache::DepIterator is_conflicted(const pkgCache::VerIterator &ver,
+				    aptitudeDepCache &cache)
+{
+  if(ver.end())
+    return pkgCache::DepIterator();
+
+  pkgCache::PkgIterator parentPkg = ver.ParentPkg();
+
+  // Look for forward conflicts:
+  for(pkgCache::DepIterator dep = ver.DependsList(); !dep.end(); ++dep)
+    if(dep->Type == pkgCache::Dep::Conflicts ||
+       dep->Type == pkgCache::Dep::DpkgBreaks)
+    {
+      // Look for direct conflicts:
+      pkgCache::VerIterator depTargetPkgInstallVer(install_version(dep.TargetPkg(),
+								   cache));
+	if(dep.TargetPkg() != parentPkg &&
+	   !depTargetPkgInstallVer.end() &&
+	   _system->VS->CheckDep(depTargetPkgInstallVer.VerStr(),
+				 dep->CompareOp,
+				 dep.TargetVer()))
+	return dep;
+
+      // Look for virtual conflicts:
+      for(pkgCache::PrvIterator prv = dep.TargetPkg().ProvidesList();
+	  !prv.end(); ++prv)
+	{
+	  if(prv.OwnerPkg() != parentPkg &&
+	     install_version(prv.OwnerPkg(), cache) == prv.OwnerVer() &&
+	     _system->VS->CheckDep(prv.ProvideVersion(),
+				   dep->CompareOp,
+				   dep.TargetVer()))
+	    return dep;
+	}
+    }
+
+  // Look for reverse conflicts:
+
+  // Look for direct reverse conflicts:
+  for(pkgCache::DepIterator dep = parentPkg.RevDependsList();
+      !dep.end(); ++dep)
+    {
+      if(dep->Type == pkgCache::Dep::Conflicts ||
+	 dep->Type == pkgCache::Dep::DpkgBreaks)
+	{
+	  if(dep.ParentPkg() != parentPkg &&
+	     install_version(dep.ParentPkg(), cache) == dep.ParentVer() &&
+	     _system->VS->CheckDep(ver.VerStr(),
+				   dep->CompareOp,
+				   dep.TargetVer()))
+	    return dep;
+	}
+    }
+
+  // Look for indirect reverse conflicts: that is, things that
+  // conflict with a package that this version provides.
+  for(pkgCache::PrvIterator prv = ver.ProvidesList();
+      !prv.end(); ++prv)
+    {
+      for(pkgCache::DepIterator dep = prv.ParentPkg().RevDependsList();
+	  !dep.end(); ++dep)
+	{
+	  if(dep->Type == pkgCache::Dep::Conflicts ||
+	     dep->Type == pkgCache::Dep::DpkgBreaks)
+	    {
+	      if(dep.ParentPkg() != parentPkg &&
+		 install_version(dep.ParentPkg(), cache) == dep.ParentVer() &&
+		 _system->VS->CheckDep(prv.ProvideVersion(),
+				       dep->CompareOp,
+				       dep.TargetVer()))
+		return dep;
+	    }
+	}
+    }
+
+  return pkgCache::DepIterator(cache, 0, (pkgCache::Version *)0);
+}
+
 /** \return \b true if d1 subsumes d2; that is, if one of the
  *  following holds:
  *
