@@ -2,6 +2,9 @@
 
 #include "aptitude_resolver_universe.h"
 
+#include "config_signal.h"
+
+#include <generic/problemresolver/problemresolver.h>
 #include <generic/problemresolver/solution.h>
 
 #include <algorithm>
@@ -9,7 +12,14 @@
 
 #include <apt-pkg/error.h>
 
+#include <cwidget/generic/util/ssprintf.h>
+
+#include <aptitude.h>
+#include <loggers.h>
+
 using namespace std;
+using aptitude::Loggers;
+using cwidget::util::ssprintf;
 
 static inline
 bool ver_disappeared(const pkgCache::VerIterator ver)
@@ -72,6 +82,8 @@ static bool empty_conflict(const pkgCache::DepIterator &dep,
 
 string aptitude_resolver_version::get_name() const
 {
+  pkgCache::VerIterator ver(get_ver());
+  pkgCache::PkgIterator pkg(get_pkg());
   if(!ver.end())
     {
       // If there are two distinct package files with the same
@@ -298,8 +310,6 @@ void aptitude_resolver_version::dep_iterator::normalize()
   // If we ran out of deps, we're done!
 }
 
-
-
 void aptitude_resolver_dep::solver_iterator::normalize()
 {
   if(!is_conflict(dep_lst->Type))
@@ -412,12 +422,14 @@ void aptitude_resolver_dep::solver_iterator::normalize()
 
 bool aptitude_resolver_dep::solved_by(const aptitude_resolver_version &v) const
 {
+  pkgCache::DepIterator start_iter(cache->GetCache(), const_cast<pkgCache::Dependency *>(start));
+
   // First check for moving the source.
-  if(v.get_pkg() == const_cast<pkgCache::DepIterator &>(start).ParentPkg() && v.get_ver() != const_cast<pkgCache::DepIterator &>(start).ParentVer())
+  if(v.get_pkg() == start_iter.ParentPkg() && v.get_ver() != start_iter.ParentVer())
     return true;
 
   // Now check each of the members of the OR group.
-  pkgCache::DepIterator d = start;
+  pkgCache::DepIterator d = start_iter;
 
   if(!is_conflict(start->Type))
     {
@@ -447,7 +459,7 @@ bool aptitude_resolver_dep::solved_by(const aptitude_resolver_version &v) const
 	    return false;
 	}
     }
-  else if(prv.end())
+  else if(prv == NULL)
     {
       if(d.TargetPkg() != v.get_pkg())
 	return false;
@@ -463,11 +475,14 @@ bool aptitude_resolver_dep::solved_by(const aptitude_resolver_version &v) const
     }
   else
     {
+      pkgCache::PrvIterator prv_iter(cache->GetCache(), const_cast<pkgCache::Provides *>(prv),
+				     (pkgCache::Package *)0);
+
       // Only other versions of the provider can solve this.
-      if(v.get_pkg() != const_cast<pkgCache::PrvIterator &>(prv).OwnerPkg())
+      if(v.get_pkg() != prv_iter.OwnerPkg())
 	return false;
       else
-	return v.get_ver() != const_cast<pkgCache::PrvIterator &>(prv).OwnerVer();
+	return v.get_ver() != prv_iter.OwnerVer();
     }
 }
 
@@ -508,7 +523,7 @@ aptitude_resolver_version aptitude_resolver_dep::solver_iterator::operator*() co
   eassert(!end());
 
   if(!ver_lst.end())
-    return aptitude_resolver_version(ver_lst.ParentPkg(),ver_lst,cache);
+    return aptitude_resolver_version::make_install(ver_lst, cache);
   else // In this case we're trying to remove some package or other.
     {
       if(!is_conflict(dep_lst->Type))
@@ -516,12 +531,12 @@ aptitude_resolver_version aptitude_resolver_dep::solver_iterator::operator*() co
 	  // Assume this because otherwise end() should be true.
 	  eassert(!prv_lst.end());
 
-	  return aptitude_resolver_version(const_cast<pkgCache::PrvIterator &>(prv_lst).OwnerPkg(),const_cast<pkgCache::PrvIterator &>(prv_lst).OwnerVer(),cache);
+	  return aptitude_resolver_version::make_install(const_cast<pkgCache::PrvIterator &>(prv_lst).OwnerVer(), cache);
 	}
       else if(!prv_lst.end())
-	return aptitude_resolver_version(const_cast<pkgCache::PrvIterator &>(prv_lst).OwnerPkg(),ver_lst,cache);
+	return aptitude_resolver_version::make_removal(const_cast<pkgCache::PrvIterator &>(prv_lst).OwnerPkg(), cache);
       else
-	return aptitude_resolver_version(const_cast<pkgCache::DepIterator &>(dep_lst).TargetPkg(),ver_lst,cache);
+	return aptitude_resolver_version::make_removal(const_cast<pkgCache::DepIterator &>(dep_lst).TargetPkg(), cache);
     }
 }
 
@@ -752,12 +767,17 @@ aptitude_universe::broken_dep_iterator &aptitude_universe::broken_dep_iterator::
   return *this;
 }
 
+std::ostream &operator<<(ostream &out, const aptitude_resolver_package &p)
+{
+  return out << p.get_name();
+}
+
 std::ostream &operator<<(ostream &out, const aptitude_resolver_version &v)
 {
   return out << v.get_package().get_name() << " " << v.get_name();
 }
 
-std::ostream &operator<<(ostream &out, aptitude_resolver_dep d)
+std::ostream &operator<<(ostream &out, const aptitude_resolver_dep &d)
 {
   std::vector<aptitude_resolver_version> solvers;
   for(aptitude_resolver_dep::solver_iterator i=d.solvers_begin(); !i.end(); ++i)
@@ -766,7 +786,7 @@ std::ostream &operator<<(ostream &out, aptitude_resolver_dep d)
   generic_solution<aptitude_universe>::ver_name_lt lt;
   sort(solvers.begin(), solvers.end(), lt);
 
-  out << d.get_source() << " -> {";
+  out << d.get_source() << (d.is_soft() ? " -S> {" : " -> {");
 
   for(std::vector<aptitude_resolver_version>::const_iterator i = solvers.begin();
       i != solvers.end(); ++i)
@@ -779,3 +799,114 @@ std::ostream &operator<<(ostream &out, aptitude_resolver_dep d)
 
   return out;
 }
+
+std::ostream &operator<<(std::ostream &out, const cfg_level &level)
+{
+  if(level.get_is_discard())
+    out << "discard";
+  else
+    out << level.get_level();
+
+  return out;
+}
+
+cfg_level aptitude_universe::parse_level(const std::string &s)
+{
+  typedef generic_problem_resolver<aptitude_universe> aptitude_resolver;
+  if(s == "maximum")
+    return cfg_level::make_level(cost_limits::maximum_level);
+  else if(s == "minimum" || s == "")
+    return cfg_level::make_level(cost_limits::minimum_level);
+  else if(s == "conflict" || s == "discard")
+    return cfg_level::make_conflict();
+  else
+    {
+      char *endptr;
+      int n = static_cast<int>(strtol(s.c_str(), &endptr, 0));
+      if(*endptr != '\0')
+	{
+	  std::string msg(ssprintf(N_("Invalid safety level \"%s\" (not \"discard\", \"maximum\", \"minimum\", or an integer)."), s.c_str()));
+	  LOG_ERROR(Loggers::getAptitudeResolverCosts(), msg);
+	  _error->Error("%s", _(msg.c_str()));
+	  return cfg_level::make_level(cost_limits::minimum_level);
+	}
+      else
+	return cfg_level::make_level(n);
+    }
+}
+
+cfg_level aptitude_universe::parse_levels(const std::string &level1,
+                                          const std::string &level2,
+                                          cfg_level default_level)
+{
+  if(level1.empty() && level2.empty())
+    return default_level;
+  else if(level1.empty())
+    return parse_level(level2);
+  else if(level2.empty())
+    return parse_level(level1);
+  else
+    return std::max<cfg_level>(parse_level(level1), parse_level(level2));
+}
+
+cfg_level aptitude_universe::get_safe_level()
+{
+  return
+    parse_levels(aptcfg->Find(PACKAGE "::ProblemResolver::Safe-Level", ""),
+                 aptcfg->Find(PACKAGE "::ProblemResolver::Safe-Tier", ""),
+                 cfg_level::make_level(10000));
+}
+
+cfg_level aptitude_universe::get_keep_all_level()
+{
+  return parse_levels(aptcfg->Find(PACKAGE "::ProblemResolver::Keep-All-Level", ""),
+                      aptcfg->Find(PACKAGE "::ProblemResolver::Keep-All-Tier", ""),
+                      cfg_level::make_level(20000));
+}
+
+cfg_level aptitude_universe::get_remove_level()
+{
+  return parse_levels(aptcfg->Find(PACKAGE "::ProblemResolver::Remove-Level", ""),
+                      aptcfg->Find(PACKAGE "::ProblemResolver::Remove-Tier", ""),
+                      cfg_level::make_level(10000));
+}
+
+cfg_level aptitude_universe::get_break_hold_level()
+{
+  return parse_levels(aptcfg->Find(PACKAGE "::ProblemResolver::Break-Hold-Level", ""),
+                      aptcfg->Find(PACKAGE "::ProblemResolver::Break-Hold-Tier", ""),
+                      cfg_level::make_level(40000));
+}
+
+cfg_level aptitude_universe::get_non_default_level()
+{
+  return parse_levels(aptcfg->Find(PACKAGE "::ProblemResolver::Non-Default-Level", ""),
+                      aptcfg->Find(PACKAGE "::ProblemResolver::Non-Default-Tier", ""),
+                      cfg_level::make_level(50000));
+}
+
+cfg_level aptitude_universe::get_remove_essential_level()
+{
+  return parse_levels(aptcfg->Find(PACKAGE "::ProblemResolver::Remove-Essential-Level", ""),
+                      aptcfg->Find(PACKAGE "::ProblemResolver::Remove-Essential-Tier", ""),
+                      cfg_level::make_level(60000));
+}
+
+bool aptitude_universe::is_candidate_for_initial_set(const aptitude_resolver_dep &d) const
+{
+  if(!d.is_soft())
+    return true;
+
+  if(cache == NULL)
+    return false;
+
+  pkgCache::DepIterator d2(d.get_dep());
+  while(!d2.end() && (d2->CompareOp & pkgCache::Dep::Or))
+    ++d2;
+
+  if(d2.end())
+    return true;
+
+  // Return true only if the dependency is *currently* not broken.
+  return (*cache)[d2] & pkgDepCache::DepGNow;
+}
