@@ -1,6 +1,6 @@
 // download_update_manager.cc
 //
-//   Copyright (C) 2005, 2007-2008 Daniel Burrows
+//   Copyright (C) 2005, 2007-2009 Daniel Burrows
 //
 //   This program is free software; you can redistribute it and/or
 //   modify it under the terms of the GNU General Public License as
@@ -135,7 +135,7 @@ namespace
 	++begin;
       }
 
-    _error->Warning("Unterminated quoted string in command: %s",
+    _error->Warning(_("Unterminated quoted string in command: %s"),
 		    whole_string.c_str());
     return rval;
   }
@@ -267,9 +267,9 @@ namespace
   }
 }
 
-download_manager::result
-download_update_manager::finish(pkgAcquire::RunResult res,
-				OpProgress &progress)
+void download_update_manager::finish(pkgAcquire::RunResult res,
+				     OpProgress *progress,
+				     const sigc::slot1<void, result> &k)
 {
   if(log != NULL)
     log->Complete();
@@ -277,9 +277,12 @@ download_update_manager::finish(pkgAcquire::RunResult res,
   apt_close_cache();
 
   if(res != pkgAcquire::Continue)
-    return failure;
+    {
+      k(failure);
+      return;
+    }
 
-  bool transientNetworkFailure = true;
+  bool transientNetworkFailure = false;
   result rval = success;
 
   // We need to claim that the download failed if any source failed,
@@ -307,12 +310,16 @@ download_update_manager::finish(pkgAcquire::RunResult res,
     }
 
   // Clean old stuff out
+  std::string listsdir = aptcfg->FindDir("Dir::State::lists");
   if(rval == success && !transientNetworkFailure &&
-     (fetcher->Clean(aptcfg->FindDir("Dir::State::lists")) == false ||
-      fetcher->Clean(aptcfg->FindDir("Dir::State::lists")+"partial/") == false))
+     aptcfg->FindB("APT::Get::List-Cleanup", true) == true &&
+     aptcfg->FindB("APT::List-Cleanup", true) == true &&
+     (fetcher->Clean(listsdir) == false ||
+      fetcher->Clean(listsdir + "partial/") == false))
     {
       _error->Error(_("Couldn't clean out list directories"));
-      return failure;
+      k(failure);
+      return;
     }
 
   // Rebuild the apt caches as done in apt-get.  cachefile is scoped
@@ -321,10 +328,11 @@ download_update_manager::finish(pkgAcquire::RunResult res,
   // redundant work at the command-line.
   {
     pkgCacheFile cachefile;
-    if(!cachefile.BuildCaches(progress, true))
+    if(!cachefile.BuildCaches(*progress, true))
       {
 	_error->Error(_("Couldn't rebuild package cache"));
-	return failure;
+	k(failure);
+	return;
       }
   }
 
@@ -360,7 +368,7 @@ download_update_manager::finish(pkgAcquire::RunResult res,
     }
   else
     {
-      progress.OverallProgress(0, 0, 1, _("Updating debtags database..."));
+      progress->OverallProgress(0, 0, 1, _("Updating debtags database..."));
 
       std::string debtags_options = aptcfg->Find(PACKAGE "::Debtags-Update-Options", "--local");
 
@@ -405,13 +413,13 @@ download_update_manager::finish(pkgAcquire::RunResult res,
 			  debtags.c_str(), debtags_options.c_str(), e.errmsg().c_str());
 	}
 
-      progress.Progress(1);
-      progress.Done();
+      progress->Progress(1);
+      progress->Done();
     }
 #endif
 
   if(need_forget_new || need_autoclean)
-    apt_load_cache(&progress, true);
+    apt_load_cache(progress, true);
 
   if(apt_cache_file != NULL && need_forget_new)
     {
@@ -431,6 +439,7 @@ download_update_manager::finish(pkgAcquire::RunResult res,
       post_autoclean_hook();
     }
 
-  return rval;
+  k(rval);
+  return;
 }
 
