@@ -1,6 +1,6 @@
 // pkg_grouppolicy.cc
 //
-//  Copyright 1999-2005, 2007-2008 Daniel Burrows
+//  Copyright 1999-2005, 2007-2010 Daniel Burrows
 //
 //  This program is free software; you can redistribute it and/or modify
 //  it under the terms of the GNU General Public License as published by
@@ -29,7 +29,8 @@
 
 #include <generic/apt/apt.h>
 #include <generic/apt/config_signal.h>
-#include <generic/apt/matchers.h>
+#include <generic/apt/matching/match.h>
+#include <generic/apt/matching/pattern.h>
 #include <generic/apt/pkg_hier.h>
 #include <generic/apt/tags.h>
 #include <generic/apt/tasks.h>
@@ -52,7 +53,8 @@ using namespace std;
 
 namespace cw = cwidget;
 
-namespace match = aptitude::matching;
+namespace matching = aptitude::matching;
+using cw::util::ref_ptr;
 
 pkg_grouppolicy_factory::~pkg_grouppolicy_factory()
 {
@@ -499,21 +501,24 @@ void pkg_grouppolicy_status::add_package(const pkgCache::PkgIterator &pkg,
 /*****************************************************************************/
 class pkg_grouppolicy_filter:public pkg_grouppolicy
 {
-  match::pkg_matcher *filter;
+  ref_ptr<matching::pattern> filter;
+  ref_ptr<matching::search_cache> search_info;
 
   pkg_grouppolicy *chain;
 public:
-  pkg_grouppolicy_filter(pkg_grouppolicy_factory *_chain, match::pkg_matcher *_filter,
+  pkg_grouppolicy_filter(pkg_grouppolicy_factory *_chain,
+			 const ref_ptr<matching::pattern> &_filter,
 			 pkg_signal *_sig, desc_signal *_desc_sig)
     :pkg_grouppolicy(_sig, _desc_sig),
      filter(_filter),
+     search_info(matching::search_cache::create()),
      chain(_chain->instantiate(_sig, _desc_sig))
   {
   }
 
   virtual void add_package(const pkgCache::PkgIterator &pkg, pkg_subtree *root)
   {
-    if(match::apply_matcher(filter, pkg, *apt_cache_file, *apt_package_records))
+    if(matching::get_match(filter, pkg, search_info, *apt_cache_file, *apt_package_records).valid())
       chain->add_package(pkg, root);
   }
 
@@ -529,7 +534,6 @@ pkg_grouppolicy *pkg_grouppolicy_filter_factory::instantiate(pkg_signal *_sig,
 pkg_grouppolicy_filter_factory::~pkg_grouppolicy_filter_factory()
 {
   delete chain;
-  delete filter;
 }
 
 /*****************************************************************************/
@@ -1132,10 +1136,10 @@ pkg_grouppolicy_task_factory::~pkg_grouppolicy_task_factory()
 }
 
 
-class pkg_grouppolicy_matchers : public pkg_grouppolicy
+class pkg_grouppolicy_patterns : public pkg_grouppolicy
 {
 public:
-  typedef pkg_grouppolicy_matchers_factory::match_pair match_pair;
+  typedef pkg_grouppolicy_patterns_factory::match_entry match_entry;
 
   struct subtree_pair
   {
@@ -1154,12 +1158,13 @@ public:
 private:
   pkg_grouppolicy_factory *chain;
   pkg_grouppolicy *passthrough_policy;
-  const vector<match_pair> &subgroups;
+  ref_ptr<matching::search_cache> search_info;
+  const vector<match_entry> &subgroups;
   typedef map<wstring, subtree_pair> subtree_map;
   subtree_map subtrees;
 
   wstring substitute(const wstring &s,
-		     match::pkg_match_result *res)
+		     const ref_ptr<matching::structural_match> &res)
   {
     wstring rval;
 
@@ -1215,14 +1220,14 @@ private:
 
 		--val;
 
-		if(val >= res->num_groups())
+		if(val >= res->get_num_groups())
 		  {
 		    string group_values;
-		    for(unsigned int i = 0; i<res->num_groups(); ++i)
+		    for(unsigned int i = 0; i < res->get_num_groups(); ++i)
 		      {
 			if(i > 0)
 			  group_values += ",";
-			group_values += res->group(i);
+			group_values += res->get_group(i);
 		      }
 
 		    wchar_t buf[1024];
@@ -1232,7 +1237,7 @@ private:
 		    return buf;
 		  }
 
-		rval += cw::util::transcode(res->group(val));
+		rval += cw::util::transcode(res->get_group(val));
 	      }
 	  }
       }
@@ -1240,16 +1245,17 @@ private:
     return rval;
   }
 public:
-  pkg_grouppolicy_matchers(pkg_grouppolicy_factory *_chain,
+  pkg_grouppolicy_patterns(pkg_grouppolicy_factory *_chain,
 			   pkg_signal *_sig, desc_signal *_desc_sig,
-			   const vector<match_pair> &_subgroups)
+			   const vector<match_entry> &_subgroups)
         :pkg_grouppolicy(_sig, _desc_sig),
 	 chain(_chain), passthrough_policy(NULL),
+	 search_info(matching::search_cache::create()),
 	 subgroups(_subgroups)
   {
   }
 
-  ~pkg_grouppolicy_matchers()
+  ~pkg_grouppolicy_patterns()
   {
     for(subtree_map::const_iterator i = subtrees.begin();
 	i != subtrees.end(); ++i)
@@ -1260,23 +1266,25 @@ public:
   void add_package(const pkgCache::PkgIterator &pkg,
 		   pkg_subtree *root)
   {
-    for(vector<match_pair>::const_iterator i = subgroups.begin();
+    for(vector<match_entry>::const_iterator i = subgroups.begin();
 	i != subgroups.end(); ++i)
 	{
-	  match::pkg_match_result *res = match::get_match(i->matcher, pkg, *apt_cache_file, *apt_package_records);
-	  if(res != NULL)
+	  ref_ptr<matching::structural_match> res(matching::get_match(i->pattern, pkg, search_info, *apt_cache_file, *apt_package_records));
+	  if(res.valid())
 	    {
+	      pkg_grouppolicy_factory * const local_chain =
+		i->chain != NULL ? i->chain : chain;
+
 	      if(i->passthrough)
 		{
 		  if(passthrough_policy == NULL)
-		    passthrough_policy = chain->instantiate(get_sig(),
-							    get_desc_sig());
+		    passthrough_policy = local_chain->instantiate(get_sig(),
+								  get_desc_sig());
 		  passthrough_policy->add_package(pkg, root);
 		  break;
 		}
 
 	      wstring title = substitute(i->tree_name, res);
-	      delete res;
 
 	      subtree_map::const_iterator found =
 		subtrees.find(title);
@@ -1286,8 +1294,8 @@ public:
 	      else
 		{
 		  pkg_subtree *tree = new pkg_subtree(title, L"", get_desc_sig());
-		  pkg_grouppolicy *policy = chain->instantiate(get_sig(),
-							       get_desc_sig());
+		  pkg_grouppolicy *policy = local_chain->instantiate(get_sig(),
+								     get_desc_sig());
 		  root->add_child(tree);
 		  tree->set_num_packages_parent(root);
 
@@ -1302,17 +1310,19 @@ public:
   }
 };
 
-pkg_grouppolicy *pkg_grouppolicy_matchers_factory :: instantiate(pkg_signal *sig,
+pkg_grouppolicy *pkg_grouppolicy_patterns_factory :: instantiate(pkg_signal *sig,
 								 desc_signal *_desc_sig)
 {
-  return new pkg_grouppolicy_matchers(chain, sig, _desc_sig, subgroups);
+  return new pkg_grouppolicy_patterns(chain, sig, _desc_sig, subgroups);
 }
 
-pkg_grouppolicy_matchers_factory :: ~pkg_grouppolicy_matchers_factory()
+pkg_grouppolicy_patterns_factory :: ~pkg_grouppolicy_patterns_factory()
 {
-  for(std::vector<match_pair>::const_iterator i = subgroups.begin();
+  for(std::vector<match_entry>::const_iterator i = subgroups.begin();
       i != subgroups.end(); ++i)
-    delete i->matcher;
+    {
+      delete i->chain;
+    }
   delete chain;
 }
 
@@ -1324,15 +1334,15 @@ class pkg_grouppolicy_tag : public pkg_grouppolicy
 {
   pkg_grouppolicy_factory *chain;
 
-  string facet;
+  string match_facet;
 
   typedef std::map<string, pair<pkg_grouppolicy *, pkg_subtree *> > childmap;
 
   childmap children;
 public:
-  pkg_grouppolicy_tag(const string &_facet, pkg_grouppolicy_factory *_chain,
+  pkg_grouppolicy_tag(const string &_match_facet, pkg_grouppolicy_factory *_chain,
 		      pkg_signal *sig, desc_signal *desc_sig)
-    :pkg_grouppolicy(sig, desc_sig), chain(_chain), facet(_facet)
+    :pkg_grouppolicy(sig, desc_sig), chain(_chain), match_facet(_match_facet)
   {
   }
 
@@ -1340,8 +1350,12 @@ public:
 			   pkg_subtree *root)
   {
 #ifdef HAVE_EPT
-    typedef ept::debtags::Tag tag;
+    using aptitude::apt::get_facet_name;
+    using aptitude::apt::get_tag_long_description;
+    using aptitude::apt::get_tag_name;
+    using aptitude::apt::get_tag_short_description;
     using aptitude::apt::get_tags;
+    using aptitude::apt::tag;
 #endif
 
 #ifdef HAVE_EPT
@@ -1358,18 +1372,15 @@ public:
 	ti != tags->end(); ++ti)
       {
 #ifdef HAVE_EPT
-	const ept::debtags::Facet f = ti->facet();
-
-	const std::string thisfacet(f.name());
+        const std::string thisfacet = get_facet_name(*ti);
 
 	// Don't create items for tags that aren't in our facet.
-	if(f.name() != facet)
+	if(thisfacet != match_facet)
 	  return;
 
 	// TODO: split up by sub-facet or whatever the debtags guys
 	// call it?
-	std::string tagname;
-	ti->name(tagname);
+	std::string tagname = get_tag_name(*ti);
 
 #else // HAVE_EPT
 	tag::const_iterator j = ti->begin();
@@ -1378,7 +1389,7 @@ public:
 
 	string thisfacet = *j;
 
-	if(thisfacet != facet)
+	if(thisfacet != match_facet)
 	  continue;
 
 	++j;
@@ -1397,11 +1408,8 @@ public:
 	if(found == children.end())
 	  {
 #ifdef HAVE_EPT
-	    std::string desc;
-	    ti->longDescription(desc);
-
-	    std::string shortdesc;
-	    ti->shortDescription(shortdesc);
+	    const std::string desc = get_tag_long_description(*ti);
+	    const std::string shortdesc = get_tag_short_description(*ti);
 #else // HAVE_EPT
 	    string desc = tag_description(ti->str());
 	    string shortdesc(desc, 0, desc.find('\n'));
@@ -1481,8 +1489,8 @@ public:
 			   pkg_subtree *root)
   {
 #ifdef HAVE_EPT
-    typedef ept::debtags::Tag tag;
     using aptitude::apt::get_tags;
+    using aptitude::apt::tag;
 #endif
 
 #ifdef HAVE_EPT
@@ -1518,9 +1526,15 @@ public:
 	ti != tags->end(); ++ti)
       {
 #ifdef HAVE_EPT
-	const ept::debtags::Facet f(ti->facet());
-	std::string thisfacet(f.name());
-	std::string thistag(ti->name());
+        using aptitude::apt::get_facet_long_description;
+        using aptitude::apt::get_facet_name;
+        using aptitude::apt::get_facet_short_description;
+        using aptitude::apt::get_tag_long_description;
+        using aptitude::apt::get_tag_name;
+        using aptitude::apt::get_tag_short_description;
+
+	std::string thisfacet(get_facet_name(*ti));
+	std::string thistag(get_tag_name(*ti));
 #else // HAVE_EPT
 	tag::const_iterator j = ti->begin();
 
@@ -1548,10 +1562,10 @@ public:
 	if(facetfound == children.end())
 	  {
 #ifdef HAVE_EPT
-	    string desc(f.longDescription());
-	    string shortdesc(f.shortDescription());
+	    string desc(get_facet_long_description(*ti));
+	    string shortdesc(get_facet_short_description(*ti));
 #else // HAVE_EPT
-	    string desc = facet_description(thisfacet);
+	    string desc = get_facet_description(*ti);
 	    string shortdesc(desc, 0, desc.find('\n'));
 #endif
 
@@ -1586,8 +1600,8 @@ public:
 	if(tagfound == tagchildren->end())
 	  {
 #ifdef HAVE_EPT
-	    string desc(ti->longDescription());
-	    string shortdesc(ti->shortDescription());
+	    string desc(get_tag_long_description(*ti));
+	    string shortdesc(get_tag_short_description(*ti));
 #else // HAVE_EPT
 	    string desc = tag_description(ti->str());
 	    string shortdesc(desc, 0, desc.find('\n'));
