@@ -1,6 +1,6 @@
 // ui.cc
 //
-//   Copyright 2000-2008 Daniel Burrows <dburrows@debian.org>
+//   Copyright 2000-2009 Daniel Burrows <dburrows@debian.org>
 //
 //   This program is free software; you can redistribute it and/or
 //   modify it under the terms of the GNU General Public License as
@@ -26,6 +26,10 @@
 #include <sigc++/functors/ptr_fun.h>
 #include <sigc++/functors/mem_fun.h>
 
+#include <boost/make_shared.hpp>
+#include <boost/shared_ptr.hpp>
+#include <boost/weak_ptr.hpp>
+
 #include <apt-pkg/acquire.h>
 #include <apt-pkg/clean.h>
 #include <apt-pkg/configuration.h>
@@ -107,6 +111,7 @@
 #include "pkg_tree.h"
 #include "pkg_ver_item.h"
 #include "pkg_view.h"
+#include "safe_slot_event.h"
 #include "ui_download_manager.h"
 #include "progress.h"
 
@@ -440,9 +445,8 @@ static void do_su_to_root(string args)
     }
   std::string root_program(root_command, splitloc + 1);
 
-  temp::dir  tempdir("aptitude");
-  temp::name statusname(tempdir, "pkgstates");
-  temp::name fifoname(tempdir, "control");
+  temp::name statusname("pkgstates");
+  temp::name fifoname("control");
 
   if(mkfifo(fifoname.get_name().c_str(), 0600) != 0)
     {
@@ -465,7 +469,7 @@ static void do_su_to_root(string args)
 
   if(pid==-1)
     {
-      _error->Error("Unable to fork: %s", strerror(errno));
+      _error->Error(_("Unable to fork: %s"), strerror(errno));
       cw::toplevel::resume();
     }
   else if(pid==0) // I'm a child!
@@ -486,7 +490,7 @@ static void do_su_to_root(string args)
       if(protocol == "su")
 	{
 	  std::ostringstream cmdbuf;
-	  cmdbuf << argv0 << " -S "
+	  cmdbuf << argv0 << " --no-gui -S "
 		 << statusname.get_name() << " "
 		 << args;
 	  execl(root_program.c_str(), root_program.c_str(), "-c", cmdbuf.str().c_str(), NULL);
@@ -497,7 +501,7 @@ static void do_su_to_root(string args)
 	{
 	  std::vector<std::string> cmdlist;
 	  // Split whitespace in the input command.
-	  std::string command = root_program + " " + argv0 + " -S " + statusname.get_name() + " " + args;
+	  std::string command = root_program + " " + argv0 + " --no-gui -S " + statusname.get_name() + " " + args;
 	  std::string::const_iterator it = command.begin();
 	  while(it != command.end() && isspace(*it))
 	    ++it;
@@ -564,14 +568,14 @@ static void do_su_to_root(string args)
 	  // We have to clear these out or the cache won't reload properly (?)
 
 	  progress_ref p = gen_progress_bar();
-	  apt_reload_cache(p.unsafe_get_ref(), true, statusname.get_name().c_str());
+	  apt_reload_cache(p->get_progress().unsafe_get_ref(), true, statusname.get_name().c_str());
 	  p->destroy();
 	}
       else
 	{
 	  // Clear out our references to these objects so they get
-	  // removed.
-	  tempdir    = temp::dir();
+	  // removed, although the shutdown routine should also do
+	  // that.
 	  statusname = temp::name();
 	  fifoname   = temp::name();
 
@@ -844,7 +848,7 @@ void do_new_package_view(OpProgress &progress)
 static void do_new_package_view_with_new_bar()
 {
   progress_ref p = gen_progress_bar();
-  do_new_package_view(*p.unsafe_get_ref());
+  do_new_package_view(*p->get_progress().unsafe_get_ref());
   p->destroy();
 }
 
@@ -867,26 +871,31 @@ static void do_new_recommendations_view()
 		  _("View packages that it is recommended that you install"),
 		  _("Recommendations"));
 
-  tree->build_tree(*p.unsafe_get_ref());
+  tree->build_tree(*p->get_progress().unsafe_get_ref());
   p->destroy();
 }
 
-static void do_new_flat_view_with_new_bar()
+void do_new_flat_view(OpProgress &progress)
 {
-  progress_ref p = gen_progress_bar();
-
   pkg_grouppolicy_factory *grp = new pkg_grouppolicy_end_factory;
   pkg_tree_ref tree = pkg_tree::create("", grp);
   tree->set_limit(cw::util::transcode("!~v"));
 
   add_main_widget(make_default_view(tree,
-				    &tree->selected_signal,
-				    &tree->selected_desc_signal),
-		  _("Packages"),
-		  _("View available packages and choose actions to perform"),
-		  _("Packages"));
+                                    &tree->selected_signal,
+                                    &tree->selected_desc_signal),
+                  _("Packages"),
+                  _("View available packages and choose actions to perform"),
+                  _("Packages"));
 
-  tree->build_tree(*p.unsafe_get_ref());
+  tree->build_tree(progress);
+}
+
+// For signal connections.
+static void do_new_flat_view_with_new_bar()
+{
+  progress_ref p = gen_progress_bar();
+  do_new_flat_view(*p->get_progress().unsafe_get_ref());
   p->destroy();
 }
 
@@ -907,7 +916,7 @@ static void do_new_tag_view_with_new_bar()
 		  _("View available packages and choose actions to perform"),
 		  _("Packages"));
 
-  tree->build_tree(*p.unsafe_get_ref());
+  tree->build_tree(*p->get_progress().unsafe_get_ref());
   p->destroy();
 }
 
@@ -936,7 +945,7 @@ void do_new_hier_view(OpProgress &progress)
 static void do_new_hier_view_with_new_bar()
 {
   progress_ref p=gen_progress_bar();
-  do_new_hier_view(*p.unsafe_get_ref());
+  do_new_hier_view(*p->get_progress().unsafe_get_ref());
   p->destroy();
 }
 
@@ -1141,38 +1150,77 @@ static void maybe_show_old_tmpdir_message()
 // (er, can I disentangle this by rearranging the routines?  I think maybe I
 //  can to some degree)
 
-
-static void finish_install_run(pkgPackageManager::OrderResult res)
+namespace
 {
-  if(res != pkgPackageManager::Incomplete)
-    {
-      cerr << _("Press return to continue.\n");
-      int c = getchar();
+  void run_dpkg_with_cwidget_suspended(sigc::slot1<pkgPackageManager::OrderResult, int> f,
+				       sigc::slot1<void, pkgPackageManager::OrderResult> k)
+  {
+    cw::toplevel::suspend();
+    pkgPackageManager::OrderResult rval = f(-1);
+    
+    if(rval != pkgPackageManager::Incomplete)
+      {
+	cerr << _("Press return to continue.\n");
+	int c = getchar();
 
-      while(c != '\n'  && c != EOF)
-	c = getchar();
-    }
+	while(c != '\n'  && c != EOF)
+	  c = getchar();
+      }
 
-  // libapt-pkg likes to stomp on SIGINT and SIGQUIT.  Restore them
-  // here in the simplest possible way.
-  cw::toplevel::install_sighandlers();
+    // libapt-pkg likes to stomp on SIGINT and SIGQUIT.  Restore them
+    // here in the simplest possible way.
+    cw::toplevel::install_sighandlers();
 
-  cw::toplevel::resume();
+    cw::toplevel::resume();
+
+    k(rval);
+    return;
+  }
+}
+
+namespace
+{
+  // Note that this is only safe if it's OK to copy the thunk in a
+  // background thread (i.e., it won't be invalidated by an object being
+  // destroyed in another thread).  In the special cases where we use
+  // this it should be all right.
+  void do_post_thunk(const safe_slot0<void> &thunk)
+  {
+    cw::toplevel::post_event(new aptitude::safe_slot_event(thunk));
+  }
+
+  progress_with_destructor make_progress_bar()
+  {
+    progress_ref rval = gen_progress_bar();
+    return std::make_pair(rval->get_progress(),
+			  sigc::mem_fun(*rval.unsafe_get_ref(),
+					&progress::destroy));
+  }
 }
 
 void install_or_remove_packages()
 {
-  download_install_manager *m = new download_install_manager(false);
+  boost::shared_ptr<download_install_manager> m =
+    boost::make_shared<download_install_manager>(false, sigc::ptr_fun(&run_dpkg_with_cwidget_suspended));
 
-  m->pre_install_hook.connect(sigc::ptr_fun(&cw::toplevel::suspend));
-  m->post_install_hook.connect(sigc::ptr_fun(&finish_install_run));
   m->post_forget_new_hook.connect(package_states_changed.make_slot());
 
-  ui_download_manager *uim =
-    new ui_download_manager(m, false, false, true,
-			    _("Downloading packages"),
-			    _("View the progress of the package download"),
-			    _("Package Download"));
+  std::pair<download_signal_log *, download_list_ref>
+    download_log_pair = gen_download_progress(false, false,
+					      _("Downloading packages"),
+					      _("View the progress of the package download"),
+					      _("Package Download"));
+
+  ui_download_manager *uim = new ui_download_manager(m,
+						     download_log_pair.first,
+						     download_log_pair.second,
+						     sigc::ptr_fun(&make_progress_bar),
+						     &do_post_thunk);
+
+  download_log_pair.second->cancelled.connect(sigc::mem_fun(*uim, &ui_download_manager::aborted));
+
+  uim->download_starts.connect(sigc::bind(sigc::ptr_fun(&ui_start_download), true));
+  uim->download_stops.connect(sigc::ptr_fun(&ui_stop_download));
 
   uim->download_complete.connect(install_finished.make_slot());
   uim->start();
@@ -1292,7 +1340,7 @@ static void do_show_preview()
 		      _("Preview"));
 
       progress_ref p=gen_progress_bar();
-      active_preview_tree->build_tree(*p.unsafe_get_ref());
+      active_preview_tree->build_tree(*p->get_progress().unsafe_get_ref());
       p->destroy();
     }
   else
@@ -1442,10 +1490,11 @@ namespace
 	for(pkgCache::PkgIterator pkg = (*apt_cache_file)->PkgBegin();
 	    !pkg.end(); ++pkg)
 	  {
-	    if(pkg->Flags & pkgCache::Flag::Essential)
-	      {
-		pkgDepCache::StateCache &state = (*apt_cache_file)[pkg];
+	    pkgDepCache::StateCache &state = (*apt_cache_file)[pkg];
 
+	    if((pkg->Flags & pkgCache::Flag::Essential) &&
+	       state.Status != 2)
+	      {
 		if(state.Delete())
 		  deleted_essential.push_back(pkg);
 		if(state.InstBroken())
@@ -1592,8 +1641,13 @@ void do_package_run()
     }
 }
 
-static void lists_autoclean_msg(download_update_manager *m)
+static void lists_autoclean_msg(boost::weak_ptr<download_update_manager> m_weak)
 {
+  boost::shared_ptr<download_update_manager> m(m_weak);
+
+  if(m.get() == NULL)
+    return;
+
   cw::widget_ref msg = cw::center::create(cw::frame::create(cw::label::create(_("Deleting obsolete downloaded files"))));
   m->post_autoclean_hook.connect(sigc::mem_fun(msg.unsafe_get_ref(),
 					       &cw::widget::destroy));
@@ -1604,15 +1658,27 @@ static void lists_autoclean_msg(download_update_manager *m)
 
 void really_do_update_lists()
 {
-  download_update_manager *m = new download_update_manager;
+  boost::shared_ptr<download_update_manager> m = boost::make_shared<download_update_manager>();
   m->pre_autoclean_hook.connect(sigc::bind(sigc::ptr_fun(lists_autoclean_msg),
-					   m));
+					   boost::weak_ptr<download_update_manager>(m)));
   m->post_forget_new_hook.connect(package_states_changed.make_slot());
-  ui_download_manager *uim =
-    new ui_download_manager(m, false, true, false,
-			    _("Updating package lists"),
-			    _("View the progress of the package list update"),
-			    _("List Update"));
+
+  std::pair<download_signal_log *, download_list_ref>
+    download_log_pair = gen_download_progress(false, true,
+					      _("Updating package lists"),
+					      _("View the progress of the package list update"),
+					      _("List Update"));
+
+  ui_download_manager *uim = new ui_download_manager(m,
+						     download_log_pair.first,
+						     download_log_pair.second,
+						     sigc::ptr_fun(&make_progress_bar),
+						     &do_post_thunk);
+
+  download_log_pair.second->cancelled.connect(sigc::mem_fun(*uim, &ui_download_manager::aborted));
+
+  uim->download_starts.connect(sigc::bind(sigc::ptr_fun(&ui_start_download), false));
+  uim->download_stops.connect(sigc::ptr_fun(&ui_stop_download));
 
   uim->download_complete.connect(update_finished.make_slot());
   uim->start();
@@ -1644,7 +1710,7 @@ static void do_sweep()
   add_main_widget(cmine::create(), _("Minesweeper"), _("Waste time trying to find mines"), _("Minesweeper"));
 }
 
-static void do_clean()
+static void really_do_clean()
 {
   if(active_download)
     // Erk!  That's weird!
@@ -1669,6 +1735,22 @@ static void do_clean()
     }
 }
 
+void do_clean()
+{
+  if(getuid()==0 || !aptcfg->FindB(PACKAGE "::Warn-Not-Root", true))
+    really_do_clean();
+  else
+    {
+	  popup_widget(cw::dialogs::yesno(wrapbox(cw::text_fragment(_("Cleaning the package cache requires administrative privileges, which you currently do not have.  Would you like to change to the root account?"))),
+				       cw::util::arg(sigc::bind(sigc::ptr_fun(&do_su_to_root),
+						      "--clean-on-startup")),
+				       W_("Become root"),
+				       cw::util::arg(sigc::ptr_fun(&really_do_update_lists)),
+				       W_("Don't become root"),
+				       cw::get_style("Error")));
+    }
+}
+
 // Ok, this is, uh, weird.  Erase has to be overridden to at least
 // call unlink()
 //
@@ -1702,7 +1784,7 @@ static bool do_autoclean_enabled()
   return apt_cache_file != NULL;
 }
 
-static void do_autoclean()
+static void really_do_autoclean()
 {
   if(apt_cache_file == NULL)
     _error->Error(_("The apt cache file is not available; cannot auto-clean."));
@@ -1736,6 +1818,22 @@ static void do_autoclean()
     }
 }
 
+void do_autoclean()
+{
+  if(getuid()==0 || !aptcfg->FindB(PACKAGE "::Warn-Not-Root", true))
+    really_do_autoclean();
+  else
+    {
+	  popup_widget(cw::dialogs::yesno(wrapbox(cw::text_fragment(_("Deleting obsolete files requires administrative privileges, which you currently do not have.  Would you like to change to the root account?"))),
+				       cw::util::arg(sigc::bind(sigc::ptr_fun(&do_su_to_root),
+						      "--autoclean-on-startup")),
+				       W_("Become root"),
+				       cw::util::arg(sigc::ptr_fun(&really_do_update_lists)),
+				       W_("Don't become root"),
+				       cw::get_style("Error")));
+    }
+}
+
 static bool do_mark_upgradable_enabled()
 {
   return apt_cache_file != NULL;
@@ -1783,12 +1881,12 @@ void do_forget_new()
 static void do_reload_cache()
 {
   progress_ref p = gen_progress_bar();
-  apt_reload_cache(p.unsafe_get_ref(), true);
+  apt_reload_cache(p->get_progress().unsafe_get_ref(), true);
   p->destroy();
 }
 #endif
 
-static void start_solution_calculation(bool blocking = true);
+static void start_solution_calculation();
 
 class interactive_continuation : public resolver_manager::background_continuation
 {
@@ -1875,54 +1973,35 @@ public:
 
   void no_more_time()
   {
-    start_solution_calculation(false);
+    start_solution_calculation();
   }
 
   void interrupted()
   {
   }
 
-  void aborted(const cw::util::Exception &e)
+  void aborted(const std::string &errmsg)
   {
-    cw::toplevel::post_event(new solution_search_aborted_event(manager, e.errmsg()));
+    cw::toplevel::post_event(new solution_search_aborted_event(manager, errmsg));
   }
 };
 
-// If the current solution pointer is past the end of the solution
-// list, queue up a calculation for it in the background thread.  If
-// "blocking" is true, then the calling thread will wait for the
-// background resolver thread to finish its calculation before
-// returning from this function.  IF THE CALLING THREAD IS THE
-// BACKGROUND THREAD THIS WILL DEADLOCK IF BLOCKING IS TRUE! (see above for such a case)
-static void start_solution_calculation(bool blocking)
+void cwidget_resolver_trampoline(const sigc::slot<void> &f)
 {
-  resolver_manager::state state = resman->state_snapshot();
+  f();
+}
 
-  if(state.resolver_exists &&
-     state.selected_solution == state.generated_solutions &&
-     !state.solutions_exhausted &&
-     !state.background_thread_active &&
-     !state.background_thread_aborted)
-    {
-      const int selected = state.selected_solution;
-      const int limit = aptcfg->FindI(PACKAGE "::ProblemResolver::StepLimit", 5000);
-      const int wait_steps = aptcfg->FindI(PACKAGE "::ProblemResolver::WaitSteps", 50);
-
-      if(limit > 0)
-	{
-	  interactive_continuation * const k = new interactive_continuation(resman);
-
-	  if(blocking)
-	    resman->get_solution_background_blocking(selected, limit, wait_steps, k);
-	  else
-	    resman->get_solution_background(selected, limit, k);
-	}
-    }
+// If the current solution pointer is past the end of the solution
+// list, queue up a calculation for it in the background thread.
+static void start_solution_calculation()
+{
+  resman->maybe_start_solution_calculation(boost::make_shared<interactive_continuation>(resman),
+					   &cwidget_resolver_trampoline);
 }
 
 static void do_connect_resolver_callback()
 {
-  resman->state_changed.connect(sigc::bind(sigc::ptr_fun(&start_solution_calculation), true));
+  resman->state_changed.connect(sigc::ptr_fun(&start_solution_calculation));
   // We may have missed a signal before making the connection:
   start_solution_calculation();
   resman->state_changed.connect(sigc::ptr_fun(&cw::toplevel::update));
@@ -2572,6 +2651,9 @@ public:
 };
 typedef cw::util::ref_ptr<help_bar> help_bar_ref;
 
+/** \brief A list of global connections that must be disconnected when the UI exits. */
+std::deque<sigc::connection> global_connections;
+
 void ui_init()
 {
   cw::toplevel::init();
@@ -2579,9 +2661,9 @@ void ui_init()
 
   // The basic behavior of the package state signal is to update the
   // display.
-  package_states_changed.connect(sigc::ptr_fun(cw::toplevel::update));
+  global_connections.push_back(package_states_changed.connect(sigc::ptr_fun(cw::toplevel::update)));
 
-  consume_errors.connect(sigc::ptr_fun(check_apt_errors));
+  global_connections.push_back(consume_errors.connect(sigc::ptr_fun(check_apt_errors)));
 
   if(aptcfg->Exists(PACKAGE "::Theme"))
     load_options(PACKAGE "::Themes::"+string(aptcfg->Find(PACKAGE "::Theme"))+"::"+PACKAGE "::UI", true);
@@ -2597,17 +2679,17 @@ void ui_init()
   aptcfg->connect(string(PACKAGE "::UI::Package-Display-Format"),
 		  sigc::ptr_fun(do_setup_columns));
 
-  cache_closed.connect(sigc::ptr_fun(do_show_reload_message));
-  cache_reloaded.connect(sigc::ptr_fun(do_hide_reload_message));
-  cache_reloaded.connect(sigc::ptr_fun(do_connect_resolver_callback));
+  global_connections.push_back(cache_closed.connect(sigc::ptr_fun(do_show_reload_message)));
+  global_connections.push_back(cache_reloaded.connect(sigc::ptr_fun(do_hide_reload_message)));
+  global_connections.push_back(cache_reloaded.connect(sigc::ptr_fun(do_connect_resolver_callback)));
   if(apt_cache_file)
     {
       do_connect_resolver_callback();
       start_solution_calculation();
     }
-  cache_reload_failed.connect(sigc::ptr_fun(do_hide_reload_message));
+  global_connections.push_back(cache_reload_failed.connect(sigc::ptr_fun(do_hide_reload_message)));
 
-  cache_reloaded.connect(sigc::ptr_fun(do_connect_read_only_callbacks));
+  global_connections.push_back(cache_reloaded.connect(sigc::ptr_fun(do_connect_read_only_callbacks)));
   if(apt_cache_file)
     do_connect_read_only_callbacks();
 
@@ -2748,8 +2830,25 @@ void ui_init()
   maybe_show_old_tmpdir_message();
 }
 
+struct clear_global_connections_on_exit
+{
+  clear_global_connections_on_exit()
+  {
+  }
+
+  ~clear_global_connections_on_exit()
+  {
+    for(std::deque<sigc::connection>::iterator it = global_connections.begin();
+	it != global_connections.end(); ++it)
+      it->disconnect();
+    global_connections.clear();
+  }
+};
+
 void ui_main()
 {
+  clear_global_connections_on_exit clearer;
+
   cw::toplevel::mainloop();
 
   if(apt_cache_file &&
@@ -2757,7 +2856,7 @@ void ui_main()
      apt_cache_file->is_locked())
     {
       progress_ref p=gen_progress_bar();
-      (*apt_cache_file)->save_selection_list(*p.unsafe_get_ref());
+      (*apt_cache_file)->save_selection_list(*p->get_progress().unsafe_get_ref());
       p->destroy();
     }
 
@@ -2865,13 +2964,12 @@ static void reset_status_download()
 }
 
 std::pair<download_signal_log *,
-	  cw::widget_ref>
+	  download_list_ref>
 gen_download_progress(bool force_noninvasive,
 		      bool list_update,
 		      const wstring &title,
 		      const wstring &longtitle,
-		      const wstring &tablabel,
-		      cw::util::slot0arg abortslot)
+		      const wstring &tablabel)
 {
   download_signal_log *m=new download_signal_log;
   download_list_ref w=NULL;
@@ -2879,14 +2977,14 @@ gen_download_progress(bool force_noninvasive,
   if(force_noninvasive ||
      aptcfg->FindB(PACKAGE "::UI::Minibuf-Download-Bar", false))
     {
-      w=download_list::create(abortslot, false, !list_update);
+      w = download_list::create(false, !list_update);
       main_status_multiplex->add_visible_widget(w, true);
       active_status_download=w;
       w->destroyed.connect(sigc::ptr_fun(&reset_status_download));
     }
   else
     {
-      w=download_list::create(abortslot, true, !list_update);
+      w = download_list::create(true, !list_update);
       add_main_widget(w, title, longtitle, tablabel);
     }
 
@@ -2916,7 +3014,7 @@ gen_download_progress(bool force_noninvasive,
   m->Complete_sig.connect(sigc::mem_fun(w.unsafe_get_ref(),
 					&download_list::Complete));
 
-  return std::pair<download_signal_log *, cw::widget_ref>(m, w);
+  return std::make_pair(m, w);
 }
 
 static void do_prompt_string(const wstring &s,
@@ -2927,20 +3025,18 @@ static void do_prompt_string(const wstring &s,
   realslot();
 }
 
-std::pair<download_signal_log *, cw::widget_ref>
+std::pair<download_signal_log *, download_list_ref>
 gen_download_progress(bool force_noninvasive,
 		      bool list_update,
 		      const string &title,
 		      const string &longtitle,
-		      const string &tablabel,
-		      cw::util::slot0arg abortslot)
+		      const string &tablabel)
 {
   return gen_download_progress(force_noninvasive,
 			       list_update,
 			       cw::util::transcode(title),
 			       cw::util::transcode(longtitle),
-			       cw::util::transcode(tablabel),
-			       abortslot);
+			       cw::util::transcode(tablabel));
 }
 
 void prompt_string(const std::wstring &prompt,
