@@ -27,7 +27,6 @@ class MiscTest : public CppUnit::TestFixture
 
   CPPUNIT_TEST(testStripWS);
   CPPUNIT_TEST(testOrderlessLt);
-  CPPUNIT_TEST(test_ssprintf);
 
   CPPUNIT_TEST_SUITE_END();
 private:
@@ -95,16 +94,6 @@ private:
     CPPUNIT_ASSERT(!cmp(d, c));
     CPPUNIT_ASSERT(!cmp(e, d));
   }
-
-  void test_ssprintf()
-  {
-    // Test that inserting very long strings via ssprintf actually works.
-    std::string horriblelongthing = "abcdefghijklmnopqrstuvwxyz";
-    while(horriblelongthing.size() < 4096)
-      horriblelongthing += horriblelongthing;
-
-    CPPUNIT_ASSERT_EQUAL(horriblelongthing + " 20", ssprintf("%s %d", horriblelongthing.c_str(), 20));
-  }
 };
 
 CPPUNIT_TEST_SUITE_REGISTRATION(MiscTest);
